import mock
import os
import unittest

import requests

from smqtk.exceptions import InvalidUriError
from smqtk.representation.data_element.girder import (
    GirderDataElement,
    girder_client  # not None when GirderDataElement is usable.
)
from smqtk.tests import TEST_DATA_DIR
from smqtk.exceptions import ReadOnlyError


DATA_KITWARE_URL = 'https://data.kitware.com'


# Check to see if we have an internet connection.
internet_available = True
try:
    r = requests.get(DATA_KITWARE_URL)
    _ = r.content
except requests.ConnectionError:
    internet_available = False


def gen_response(content, status_code=200):
    resp = requests.Response()
    resp._content = content
    resp.status_code = status_code
    resp.headers['content-length'] = len(content)
    return resp


# Only perform these tests if the implementation is usable.
if GirderDataElement.is_usable():

    class TestGirderDataElement (unittest.TestCase):
        """
        Tests for the GirderDataElement plugin implementation

        """

        LOCAL_APIROOT = "http://localhost:8080/api/v1"
        EXAMPLE_ITEM_ID = '5820bbeb8d777f10f26efc2f'
        EXAMPLE_GIRDER_API_ROOT = "%s/api/v1" % DATA_KITWARE_URL
        EXAMPLE_GIRDER_FULL_URI = (
            EXAMPLE_GIRDER_API_ROOT.replace('https', 'girder')
            + '/file/%s' % EXAMPLE_ITEM_ID
        )
        EXAMPLE_PTH = os.path.join(TEST_DATA_DIR, 'Lenna.png')
        EXAMPLE_SHA512 = 'ca2be093f4f25d2168a0afebe013237efce02197bcd8a87f41f' \
                         'f9177824222a1ddae1f6b4f5caf11d68f2959d13f399ea55bd6' \
                         '77c979642d723e02eb9b5dc4d5'

        def test_new_fileId(self):
            expected_id = "some id"
            e = GirderDataElement(expected_id)
            self.assertEqual(e.file_id, expected_id)
            self.assertEqual(e.api_root, self.LOCAL_APIROOT)
            self.assertIsNone(e.api_key)
            self.assertIsNone(e.token)
            self.assertIsNone(e._content_type)
            self.assertIsInstance(e.gc, girder_client.GirderClient)

        @mock.patch('girder_client.GirderClient.authenticate')
        def test_repr(self, _mock_requests):
            expected_file_id = 'some_file id'
            expected_api_root = 'https://some.server/api/v1'
            expected_api_key = 'someKeyHere'

            e = GirderDataElement(expected_file_id, expected_api_root,
                                  expected_api_key)
            actual_repr = repr(e)

            expected_repr = "GirderDataElement{file_id: %s, " \
                            "api_root: %s, " \
                            "api_key: %s, token: }" \
                            % (expected_file_id, expected_api_root,
                               expected_api_key)
            self.assertEqual(actual_repr, expected_repr)

        def test_configuration_default(self):
            default_config = GirderDataElement.get_default_config()
            self.assertEqual(default_config,
                             {"file_id": None,
                              "api_root": self.LOCAL_APIROOT,
                              "api_key": None,
                              "token": None})

        @mock.patch('girder_client.GirderClient.authenticate')
        def test_from_config_full_constructor(self, _mock_authenticate):
            expected_file_id = '34uhki34gh2345ghjk'
            expected_api_root = 'https://some.other.server/api/v1'
            expected_api_key = '1234ghk135hlg23435'
            new_config = {
                'token': None,
                'file_id': expected_file_id,
                'api_root': expected_api_root,
                'api_key': expected_api_key,
            }
            #: :type: GirderDataElement
            e = GirderDataElement.from_config(new_config)
            self.assertEqual(e.file_id, expected_file_id)
            self.assertEqual(e.api_root, expected_api_root)
            self.assertEqual(e.get_config(), new_config)

        def test_from_config_common_partial(self):
            expected_file_id = '5hjkl1345hjk'
            expected_api_root = self.LOCAL_APIROOT
            expected_api_key = None
            #: :type: GirderDataElement
            e = GirderDataElement.from_config({'file_id': expected_file_id})
            self.assertEqual(e.file_id, expected_file_id)
            self.assertEqual(e.api_root, expected_api_root)
            self.assertEqual(e.get_config(),
                             {'file_id': expected_file_id,
                              'api_root': expected_api_root,
                              'api_key': expected_api_key,
                              'token': None})

        def test_from_uri_full_url(self):
            e = GirderDataElement.from_uri(self.EXAMPLE_GIRDER_FULL_URI)
            self.assertEqual(e.file_id, self.EXAMPLE_ITEM_ID)

        def test_from_uri_bad_tag(self):
            # Ensures we catch a bad tag in the URI, i.e., one that is neither
            # girder nor girders.
            self.assertRaises(InvalidUriError, GirderDataElement.from_uri,
                              uri='a_bad_tag')

        def test_from_uri_bad_path(self):
            # Ensures that we catch a URI that has an appropriate tag and
            # netloc, but the path does not begin with /api, so it is an invalid
            # girder API root.
            self.assertRaises(InvalidUriError, GirderDataElement.from_uri,
                              uri='girder://localhost:8080/bad/path')

        @mock.patch('girder_client.GirderClient.getFile')
        def test_content_type_no_cache(self, m_getFile):
            # Mocking such that we simulate a valid API root and an existing
            # item reference

            # Dummy requests return value
            expected_mimetype = 'some/type'
            m_getFile.return_value = {
                    'mimeType': expected_mimetype
            }

            e = GirderDataElement('foo')
            actual_type = e.content_type()
            self.assertEqual(actual_type, expected_mimetype)
            m_getFile.assert_called_once()

            # Ensure that calling content_type a second time doesn't call
            # getFile again.
            self.assertEqual(e.content_type(), expected_mimetype)
            m_getFile.assert_called_once()

        @mock.patch('girder_client.GirderClient.getFile')
        def test_get_file_model(self, m_getFile):
            # Check static expected values in model. This happens to be actual
            # return values, however we are mocking out any actual network
            # communication.
            expected_m = {
                '_id': self.EXAMPLE_ITEM_ID,
                '_modelType': 'file',
                'exts': ['png'],
                'mimeType': 'image/png',
                'name': 'Lenna.png',
                'sha512': self.EXAMPLE_SHA512,
                'size': 473831
            }

            m_getFile.return_value = expected_m

            e = GirderDataElement(self.EXAMPLE_ITEM_ID,
                                  self.EXAMPLE_GIRDER_API_ROOT)
            m = e.get_file_model()

            self.assertEqual(m['_id'], expected_m['_id'])
            self.assertEqual(m['_modelType'], expected_m['_modelType'])
            self.assertEqual(m['exts'], expected_m['exts'])
            self.assertEqual(m['mimeType'], expected_m['mimeType'])
            self.assertEqual(m['name'], expected_m['name'])
            self.assertEqual(m['sha512'], expected_m['sha512'])
            self.assertEqual(m['size'], expected_m['size'])

        @mock.patch('girder_client.GirderClient.getFile')
        def test_get_file_model_item_no_exists(self, m_getFile):
            def raise_http_error(*_, **__):
                raise girder_client.HttpError(None, None, None, None)
            m_getFile.side_effect = raise_http_error

            e = GirderDataElement('foo', self.EXAMPLE_GIRDER_API_ROOT)
            m = e.get_file_model()
            self.assertIsNone(m)

        def test_is_empty_none_model(self):
            # Uses model return, empty if no model return (no item in girder by
            # ID)
            e = GirderDataElement('someId')
            e.get_file_model = mock.MagicMock(return_value=None)
            self.assertTrue(e.is_empty())

        def test_is_empty_zero_size(self):
            # Uses model return size parameter
            e = GirderDataElement('someId')
            e.get_file_model = mock.MagicMock(return_value={'size': 0})
            self.assertTrue(e.is_empty())

        def test_is_empty_nonzero_bytes(self):
            e = GirderDataElement('someId')
            e.get_file_model = mock.MagicMock(return_value={'size': 7})
            self.assertFalse(e.is_empty())

        @mock.patch('smqtk.representation.data_element.girder.GirderDataElement'
                    '.get_file_model')
        @mock.patch('girder_client.GirderClient.getFolder')
        @mock.patch('girder_client.GirderClient.getItem')
        def test_writable(self, m_getItem, m_getFolder, _m_get_file_model):
            m_getItem.return_value = {'folderId': 'someFolderId'}
            m_getFolder.return_value = {'_accessLevel': 1}

            self.assertTrue(GirderDataElement('someId').writable())

            # Access level 0 should cause it to be unwritable
            m_getFolder.return_value = {'_accessLevel': 0}
            self.assertFalse(GirderDataElement('someId').writable())

            # A nonexistent file model should make writable return false
            gde = GirderDataElement('someId')
            gde.get_file_model = mock.MagicMock(return_value=None)
            self.assertFalse(gde.writable())

        @mock.patch('girder_client.GirderClient.uploadFileContents')
        def test_set_bytes(self, m_uploadFileContents):
            gde = GirderDataElement('someId')
            gde.writable = mock.MagicMock(return_value=True)
            gde.set_bytes(b'foo')
            m_uploadFileContents.assert_called_once()

        def test_set_bytes_non_writable(self):
            gde = GirderDataElement('someId')
            gde.writable = mock.MagicMock(return_value=False)
            self.assertRaises(ReadOnlyError, gde.set_bytes, b=None)

        def test_set_bytes_http_errors(self):
            gde = GirderDataElement('someId')
            gde.writable = mock.MagicMock(return_value=True)

            # Test access denied throws ReadOnlyError
            gde.gc.uploadFileContents = mock.MagicMock(
                side_effect=girder_client.HttpError(401, '', None, None))
            self.assertRaises(ReadOnlyError, gde.set_bytes, b=b'foo')

            # Test any other error (like a 500) re-raises the HttpError
            gde.gc.uploadFileContents = mock.MagicMock(
                side_effect=girder_client.HttpError(500, '', None, None))
            self.assertRaises(girder_client.HttpError, gde.set_bytes, b=b'foo')

        @mock.patch('girder_client.GirderClient.downloadFile')
        @mock.patch('six.BytesIO.getvalue')
        def test_get_bytes(self, m_getvalue, _m_downloadFile):
            m_getvalue.return_value = 'foo'

            e = GirderDataElement('someId')
            actual_bytes = e.get_bytes()

            self.assertEqual(actual_bytes, bytes('foo'))



"""
LICENCE
-------
Copyright 2015 by Kitware, Inc. All Rights Reserved. Please refer to
KITWARE_LICENSE.TXT for licensing information, or contact General Counsel,
Kitware, Inc., 28 Corporate Drive, Clifton Park, NY 12065.

"""

from math import acos, pi
import numpy as np


def histogram_intersection_distance(a, b):
    """
    Compute the histogram intersection distance between given histogram
    vectors or matrices ``a`` and ``b``, returning a value between ``0.0`` and
    ``1.0``. This is the inverse of intersection similarity, whereby a distance
    of  ``0.0`` means full intersection and ``1.0`` means no intersection.

    This implements non-branching formula for efficient computation.

    Input vectors ``a`` and ``b`` may be of 1 or 2 dimensions. Depending on the
    values of ``a`` and ``b``, different things may occur:

    * If both ``a`` and ``b`` are 2D matrices, they're shape must be congruent
      and the result will be an vector of distances between parallel rows.

    * If either is a 1D vector and the other is a 2D matrix, a vector of
      distances is returned between the 1D vector and each row of the 2D matrix.

    * If both ``a`` and ``b`` are 1D vectors, a floating-point scalar distance
      is returned that is the histogram intersection distance between the input
      vectors.

    :param a: Histogram or array of histograms ``a``
    :type a: numpy.core.multiarray.ndarray

    :param b: Histogram or array of histograms ``b``
    :type b: numpy.core.multiarray.ndarray

    :return: Float or array of float distance (inverse percent intersection).
    :rtype: float or numpy.core.multiarray.ndarray
    """
    # TODO: input value checks?
    # Which axis to sum on. If there is a matrix involved, its along column,
    # but if its just two arrays its along the row.
    #
    # - The following are noticeably slower:
    #       sum_axis = not (a.ndim == 1 and b.ndim == 1)
    #       sum_axis = (a.ndim > 1) | (b.ndim > 1)
    sum_axis = 1
    if a.ndim == 1 and b.ndim == 1:
        sum_axis = 0
    # TODO(john.moeller): Assuming each sums to 1, this can be sped up
    #   return np.abs(np.subtract(i, j)).sum(sum_axis) * 0.5
    return 1. - ((np.add(a, b) - np.abs(np.subtract(a, b))).sum(sum_axis) * 0.5)


def histogram_intersection_distance_fast(i, j):
    """
    Compute the histogram intersection percent relation between given 1D
    histogram vectors ``a`` and ``b``, returning a value between 0.0 and 1.0.
    0.0 means full intersection, and 1.0 means no intersection.

    This implements non-branching formula for efficient computation.

    Use of this implementations is faster when the input will only be 1D
    vectors.

    PENDING DEPRECATION: This function doesn't time much, if any, better than
        ``histogram_intersection_distance`` (via ipython %timeit using random
        input).

    :param i: Histogram ``i``
    :type i: numpy.core.multiarray.ndarray

    :param j: Histogram ``j``
    :type j: numpy.core.multiarray.ndarray

    :return: Float inverse percent intersection amount.
    :rtype: float

    """
    return 1.0 - ((i + j - np.abs(i - j)).sum() * 0.5)


def euclidean_distance(i, j):
    """
    Compute euclidean distance between two N-dimensional point vectors.

    :param i: Vector i
    :type i: np.ndarray

    :param j: Vector j
    :type j: np.ndarray

    :return: Float distance.
    :rtype: float | np.ndarray

    """
    sum_axis = 1
    if i.ndim == 1 and j.ndim == 1:
        sum_axis = 0
    # noinspection PyTypeChecker
    return np.sqrt(np.square(i - j).sum(sum_axis))


def cosine_similarity(i, j):
    """
    Angular similarity between vectors i and j. Results in a value between 1,
    where i and j are exactly the same, to -1, meaning exactly opposite. 0
    indicates orthogonality. Negative values will only be returned if input
    vectors can have negative values.

    See: http://en.wikipedia.org/wiki/Cosine_similarity

    :param i: Vector i
    :type i: numpy.core.multiarray.ndarray

    :param j: Vector j
    :type j: numpy.core.multiarray.ndarray

    :return: Float similarity.
    :rtype: float

    """
    # numpy.linalg.norm is Frobenius norm (vector magnitude)
    # return numpy.dot(i, j) / (numpy.linalg.norm(i) * numpy.linalg.norm(j))

    # speed optimization, numpy.linalg.norm can be a bottleneck
    # TODO(john.moeller): This can be made array-friendly
    #   np.multiply(i, j).sum(axis) / (np.linalg.norm(i, i_axis) *
    #       np.linalg.norm(j, j_axis))
    return np.dot(i, j) / (np.sqrt(i.dot(i)) * np.sqrt(j.dot(j)))


def cosine_distance(i, j, pos_vectors=True):
    """
    Cosine similarity converted into angular distance.

    See: https://en.wikipedia.org/wiki/Cosine_similarity, section
    "Angular distance and similarity".

    :param i: Vector i
    :type i: numpy.core.multiarray.ndarray

    :param j: Vector j
    :type j: numpy.core.multiarray.ndarray

    :param pos_vectors: If we expect vector elements to always be positive.
        Default value is True (common case).
    :type pos_vectors: bool

    :return: Float distance between [0, 1] range.
    :rtype: float

    """
    sim = max(-1.0, min(cosine_similarity(i, j), 1))
    return (1 + bool(pos_vectors)) * acos(sim) / pi


def hamming_distance(i, j):
    """
    Return the hamming distance between the two given pythonic integers, or the
    number of places where the bits differ.

    **Note:** *We say "pythonic" integer here because this function has no cap
    on the number of bits used to represent said integer. This function will
    execute correctly regardless whether i/j is 32 bits or 512 bits, etc."

    :param i: First integer.
    :type i: int | long
    :param j: Second integer.
    :type j: int | long

    :return: Integer hamming distance between the two values.
    :rtype: int | long

    """
    # TODO: Find something better than this?
    return bin(i ^ j).count('1')



"""
LICENCE
-------
Copyright 2013-2015 by Kitware, Inc. All Rights Reserved. Please refer to
KITWARE_LICENSE.TXT for licensing information, or contact General Counsel,
Kitware, Inc., 28 Corporate Drive, Clifton Park, NY 12065.
"""

import logging
import multiprocessing
import numpy as np

from smqtk.utils import ReadWriteLock
from smqtk.utils import SimpleTimer
from smqtk.utils.distance_kernel import DistanceKernel


class FeatureMemory (object):
    """
    Class for encapsulating and managing feature and kernel matrices for
    different feature types
    """

    @classmethod
    def construct_from_files(cls, id_vector_file, bg_flags_file,
                             feature_mat_file, kernel_mat_file, rw_lock=None):
        """ Initialize FeatureMemory object from file sources.

        :param id_vector_file: File containing the numpy.savetxt(...) output of
            clip ID values in the order in which they associate to the rows of
            the kernel matrix.
        :type id_vector_file: str
        :param feature_mat_file: File containing the kernel matrix as saved by
            numpy.save(...) (saved as an ndarray, converted to matrix on load).
        :type feature_mat_file: str
        :param kernel_mat_file: File containing the kernel matrix as saved by
            numpy.save(...) (saved as an ndarray, converted to matrix on load).
        :type kernel_mat_file: str
        :param bg_flags_file: Optional file containing output of
            numpy.savetxt(...) where each index maps a row index of the kernel
            to whether or not the associated clip ID should be considered a
            background video or not.
        :type bg_flags_file: str
        :param rw_lock: Optional ReadWriteLock for this instance to use. If not
            provided, we will create our own.
        :type rw_lock: None or ReadWriteLock
        :return: Symmetric FeatureMemory constructed with the data provided in
            the provided files.
        :rtype: FeatureMemory

        """
        clip_ids = np.array(np.load(id_vector_file))
        bg_flags = np.array(np.load(bg_flags_file))
        # noinspection PyCallingNonCallable
        feature_mat = np.matrix(np.load(feature_mat_file))
        # noinspection PyCallingNonCallable
        kernel_mat = np.matrix(np.load(kernel_mat_file))

        bg_clips = set([clip_ids[i]
                        for i, f in enumerate(bg_flags)
                        if f])

        return FeatureMemory(clip_ids, bg_clips, feature_mat, kernel_mat,
                             rw_lock=rw_lock)

    @property
    def _log(self):
        return logging.getLogger('.'.join([self.__module__,
                                           self.__class__.__name__]))

    def __init__(self, id_vector, bg_clip_ids, feature_mat, kernel_mat,
                 rw_lock=None):
        """ Initialize this FeatureMemory object

        This class must be used with numpy ndarray and matrix classes for shared
        memory purposes.

        NOTE: Arrays and matrices given here must own their data! This is
        currently required in order to resize them later when updating with new
        feature vectors. A ValueError will be thrown if an given array/matrix
        does not own its data.

        TODO: Allow kernel matrix to be optional, causing it to be built from
        the provided feature matrix (not a recommended action).

        :param id_vector: (numpy) Array of clip IDs. This is used as the map
            from an index position to the clip ID its associated with in the
            kernel and distance kernel matrices.
        :type id_vector: ndarray of int
        :param bg_clip_ids: Set of clip IDs that are to be treated as background
            clip IDs.
        :type bg_clip_ids: set of int
        :param feature_mat: (numpy) Matrix of features for clip IDs. Features
            should be stored vertically, i.e. Each row is a feature for a
            particular clip ID (id_vector being the index-to-clipID map).
        :type feature_mat: matrix of double
        :param kernel_mat: (numpy) Matrix detailing the distances between
            feature vectors. This must be a square, symmetric matrix.
        :type kernel_mat: matrix of double
        :param rw_lock: Optional ReadWriteLock for this instance to use. If not
            provided, we will create our own.
        :type rw_lock: None or ReadWriteLock

        """
        # assert isinstance(id_vector, (ndarray, ArrayProxy)), \
        #     "ID vector not given as a numpy.ndarray!"
        assert isinstance(bg_clip_ids, (set, frozenset)), \
            "Background ID vector not a numpy.ndarray!"
        # assert isinstance(feature_mat, (matrix, MatrixProxy)), \
        #     "Kernel matrix not a numpy.matrix!"
        # assert isinstance(kernel_mat, (matrix, MatrixProxy)), \
        #     "Distance kernel not a numpy.matrix!"

        # noinspection PyUnresolvedReferences
        # -> base IS a member of the matrix class...
        if id_vector.base is not None:
            raise ValueError("Given ``id_vector`` does not own its data! It "
                             "will not be transformable later.")
        elif feature_mat.base is not None:
            raise ValueError("Given ``feature_mat`` does not own its data! It "
                             "will not be transformable later.")
        elif kernel_mat.base is not None:
            raise ValueError("Given ``kernel_mat`` does not own its data! It "
                             "will not be transformable later.")

        # The kernel should be square and should be the same size as the feature
        # matrix's number or rows (unique stored clip features).
        if not (kernel_mat.shape[0] == kernel_mat.shape[1] ==
                feature_mat.shape[0]):
            raise ValueError("The distance kernel matrix provided is either "
                             "misshapen or conflicts with the dimensions of "
                             "the provided feature matrix. (kernel matrix "
                             "shape: %s, num feature vectors: %d"
                             % (kernel_mat.shape, feature_mat.shape[0]))

        self._log.debug("Lock given: %s", rw_lock)
        if rw_lock:
            assert isinstance(rw_lock, ReadWriteLock), \
                "Not given a value ReadWriteLock instance!"
            self._rw_lock = rw_lock
        else:
            self._log.debug("Falling back on bad lock given (given: %s)",
                            type(rw_lock))
            self._rw_lock = ReadWriteLock()

        self._id_vector = id_vector
        self._bg_clip_ids = bg_clip_ids
        self._feature_mat = feature_mat
        self._kernel_mat = kernel_mat

        # Helper structure mapping clipIDs to their row index
        self._cid2idx_map = dict((cid, idx) for idx, cid
                                 in enumerate(self._id_vector))

    @staticmethod
    def _histogram_intersection_distance(a, b):
        """
        Calculates distance between two vectors using histogram intersection.

        Non-branching version of the histogram intersection algorithm.

        :param a: A vector in array form.
        :type a: ndarray
        :param b: A vector in array form.
        :type b: ndarray

        :return: Histogram Intersection (HI) distance scalar
        :rtype: double

        """
        # noinspection PyUnresolvedReferences
        return (a + b - np.abs(a - b)).sum() * 0.5

    def get_ids(self):
        """
        NOTE: NOT THREAD SAFE. Use the returned structure only in conjunction
        with this object's lock when in a parallel environment to prevent
        possible memory corruption.

        :return: Ordered vector of clip IDs along the row-edge of this object's
            feature matrix and along both edges of the kernel matrix.
        :rtype: numpy.core.multiarray.ndarray

        """
        return self._id_vector

    def get_bg_ids(self):
        """
        NOTE: NOT THREAD SAFE. Use the returned structure only in conjunction
        with this object's lock when in a parallel environment to prevent
        possible memory corruption.

        :return: Ordered vector of clip IDs that we are treating as background
            clips.
        :rtype: ndarray

        """
        return frozenset(self._bg_clip_ids)

    def get_feature_matrix(self):
        """
        NOTE: NOT THREAD SAFE. Use the returned structure only in conjunction
        with this object's lock when in a parallel environment to prevent
        possible memory corruption.

        :return: Matrix recording feature vectors for a feature type. See the
            id vector for row-wise index-to-clipID association.
        :rtype: numpy.matrixlib.defmatrix.matrix

        """
        return self._feature_mat

    def get_kernel_matrix(self):
        """
        NOTE: NOT THREAD SAFE. Use the returned structure only in conjunction
        with this object's lock when in a parallel environment to prevent
        possible memory corruption.

        :return: Symmetric matrix detailing the distances between any two clip
            ID features. Distances are computed via histogram intersection.
        :rtype: matrix

        """
        return self._kernel_mat

    def get_lock(self):
        """
        :return: a reference to this object's read/write lock.
        :rtype: ReadWriteLock

        """
        return self._rw_lock

    def get_distance_kernel(self):
        """
        DistanceKernel object constructed from this feature's current state.

        :return: This feature distance kernel.
        :rtype: DistanceKernel

        """
        with self._rw_lock.read_lock():
            return DistanceKernel(self._id_vector, self._id_vector,
                                  self._kernel_mat, self._bg_clip_ids,
                                  self._rw_lock)

    def get_feature(self, *clip_id_or_ids):
        """
        Return the a matrix where each row is the feature vector for one or more
        clip IDs. The given list of clip IDs given acts as the index-to-clipID
        map for the returned matrix's rows. If repeat clip IDs are provided in
        the input, there will be repeat feature vectors in the returned matrix.

        Raises ValueError if the given clip ID is not represented in the current
        matrix.

        :param clip_id_or_ids: One or more integer clip IDs to retrieve the
            feature vectors for.
        :type clip_id_or_ids: tuple of int

        :return: NxM matrix, where N is the number of clip IDs requested and M
            is the length of a feature vector for this vector.
        :rtype: np.matrix

        """
        assert all(isinstance(e, int) for e in clip_id_or_ids), \
            "Not given an integer or a valid iterable over integers!"

        with self._rw_lock.read_lock():
            # rows = num of IDs given, cols = width of feature matrix
            with SimpleTimer("Allocating return matrix", self._log.debug):
                # noinspection PyUnresolvedReferences
                # -> matrix class DOES have ``dtype`` property...
                ret_mat = matrix(ndarray((len(clip_id_or_ids),
                                          self._feature_mat.shape[1]),
                                         self._feature_mat.dtype))
            for i, cid in enumerate(clip_id_or_ids):
                feature_idx = self._cid2idx_map[cid]
                ret_mat[i, :] = self._feature_mat[feature_idx, :]
            return ret_mat

    # noinspection PyUnresolvedReferences,PyCallingNonCallable
    def update(self, clip_id, feature_vec=None, is_background=False,
               timeout=None):
        """
        Update this feature with a feature vector associated with a clip ID. If
        clip ID is already in the feature matrix, we replace the current vector
        with the given one.

        Either way, the distance kernel is updated with either a new row/column,
        or updating relevant slots in the existing distance kernel.

        :raise ValueError: if the given feature vector is not compatible with
            our feature vector.
        :raise RuntimeError: If a timeout is given and the underlying write lock
            doesn't acquire in that amount of time.

        :param clip_id: The ID of the clip the given ``feature_vec`` represents.
        :type clip_id: int
        :param feature_vec: Feature vector associated to the given clip ID.
        :type feature_vec: ndarray
        :param is_background: Flag declaring that this clip ID represents a
            background feature.
        :type is_background: bool
        :param timeout: Timeout seconds for the underlying write lock to acquire
            before a RuntimeError is thrown.
        :type timeout: None or int or float

        """
        with self._rw_lock.write_lock(timeout):
            clip_id = int(clip_id)
            if feature_vec is not None and \
                    not (feature_vec.ndim == 1
                         and len(feature_vec) == self._feature_mat.shape[1]):
                raise ValueError("Given feature vector not compatible "
                                 "(dimensionality or length does not match)")

            # Update the given feature vector and kernel distances
            if self._cid2idx_map.get(clip_id, None) is not None:
                # In all cases, update the background status of the clip
                if is_background:
                    self._bg_clip_ids.add(clip_id)
                else:
                    self._bg_clip_ids.discard(clip_id)

                # If we were given a new feature vector, update entries
                if feature_vec is not None:
                    idx = self._cid2idx_map[clip_id]
                    self._feature_mat[idx] = feature_vec
                    new_dist = np.mat(tuple(
                        self._histogram_intersection_distance(feature_vec, fv)
                        for fv in self._feature_mat
                    ))
                    self._kernel_mat[idx, :] = new_dist
                    self._kernel_mat[:, idx] = new_dist

            # Given a new feature to add.
            else:
                if feature_vec is None:
                    raise ValueError("Update given a new clip ID, but no "
                                     "feature vector provided.")

                # Update internal feature matrix with added vector
                self._cid2idx_map[clip_id] = self._id_vector.size
                self._id_vector.resize((self._id_vector.size + 1,),
                                       refcheck=False)
                self._id_vector[-1] = clip_id

                if is_background:
                    self._bg_clip_ids.add(clip_id)

                # noinspection PyUnresolvedReferences
                if self._feature_mat.base is not None:
                    raise RuntimeError("Feature matrix does not own its data")
                # Since we're only adding a new row, this resize does not affect
                # the positioning of the existing data.
                # noinspection PyUnresolvedReferences
                self._feature_mat.resize((self._feature_mat.shape[0] + 1,
                                          self._feature_mat.shape[1]),
                                         refcheck=False
                                         )
                self._feature_mat[-1, :] = feature_vec

                # Need to add a new row AND column to the distance kernel.
                if self._kernel_mat.base is not None:
                    raise RuntimeError("kernel matrix does not own its data")
                assert self._kernel_mat.shape[0] == self._kernel_mat.shape[1], \
                    "kernel matrix is not symmetric for some reason???"
                # noinspection PyPep8Naming
                # -> because I like ``N`` better...
                N = self._kernel_mat.shape[0]
                kernel_copy = np.matrix(self._kernel_mat)
                self._kernel_mat.resize((N+1, N+1), refcheck=False)
                self._kernel_mat[:N, :N] = kernel_copy
                del kernel_copy

                # Computing new feature distance (histogram intersection). Only
                # need to compute this once because of HI being being
                # commutative and the kernel matrix being symmetric.
                dist_vec = np.mat(tuple(
                    self._histogram_intersection_distance(feature_vec, fv)
                    for fv in self._feature_mat
                ))
                self._kernel_mat[-1, :] = dist_vec
                self._kernel_mat[:, -1] = dist_vec.T


class FeatureMemoryMap (object):
    """ Map different feature types to their own FeatureMemory object
    """
    # Basically a pass-through for all the functions in FeatureMemory, but with
    # and additional heading parameter of feature type.

    def __init__(self):
        self._map_lock = multiprocessing.RLock()
        #: :type: dict[str, FeatureMemory]
        self._feature2memory = {}

    def get_feature_types(self):
        """ Get available feature types in this map.

        :return: Tuple of string names of all features initialize in this map
        :rtype: tuple of str

        """
        with self._map_lock:
            return list(self._feature2memory.keys())

    def initialize(self, feature_type, id_vector, bg_clip_ids_vector,
                   feature_mat, kernel_mat, rw_lock=None):
        """ Initialize a feature type within this map

        :raise KeyError: When the given feature_type is already present in the
            map (requires a removal first).
        :raise ValueError: When there is an issue with one or more of the
            provided input data elements.

        :param feature_type: The name assigned to the feature type
        :param id_vector: Numpy array mapping indices to a clip ID.
        :param bg_clip_ids_vector: Numpy array of integers detailing the clip
            IDs that are to be considered background.
        :param feature_mat: Numpy matrix of all clip feature vectors. This
            should be an order 'C' matrix with features stacked vertically. The
            ``id_vector`` maps row indices to the clip ID the feature
            represents.
        :param kernel_mat: Pre-computed distance kernel
        :param rw_lock: Optional read-write lock to manually use with underlying
            feature memory construct.  Otherwise we create our own.

        """
        with self._map_lock:
            # KeyError on repeat feature_type key, require removal first
            if feature_type in self._feature2memory:
                raise KeyError("Key '%s' already present in our mapping. "
                               "Please remove first before initializing."
                               % feature_type)
            self._feature2memory[feature_type] = \
                FeatureMemory(id_vector, bg_clip_ids_vector, feature_mat,
                              kernel_mat, rw_lock)

    def initialize_from_files(self, feature_type, id_vector_file, bg_flags_file,
                              feature_mat_file, kernel_mat_file, rw_lock=None):
        """ Initialize a feature type within this map from file resources.

        Files pointed to must be of the following formats:

            - id_vector_file:
                File resulting from a numpy.savetxt() of a one-dimensional
                array, mapping index position with an integer clip ID. This
                should correlate to the clip IDs of the row-major features
                stored in feature_mat_file.

            - bg_flags_file:
                File resulting from a numpy.savetxt() of a one-dimensional
                array, mapping index position with whether that clip should be
                treated as a background video or not. This should corrolate with
                features in the same way as the id_id_vector_file.

            - feature_mat_file:
                File resulting from a numpy.save() of an ndarray. This will be
                loaded in as a matrix. This should be the initial NxD feature
                matrix for this feature type.

            - kernel_mat_file:
                File resulting from a numpy.save() of an ndarray. This will be
                loaded in as a matrix. This should be the computed NxN distance
                kernel for this feature type.

        :param feature_type: The name assigned to the feature type.
        :type id_vector_file: str
        :type bg_flags_file: str
        :type feature_mat_file: str
        :type kernel_mat_file: str
        :param rw_lock: Optionally specified ReadWriteLock instance to use with
            the underlying FeatureMemory and DistanceKernel objects associated
            with this feature type (not recommended)
        :type rw_lock: ReadWriteLock

        """
        with self._map_lock:
            # even though this happens in the initialize() call we make here,
            # we would like to short circuit before loading data if we can.
            if feature_type in self._feature2memory:
                raise KeyError("Key '%s' already present in our mapping. "
                               "Please remove first before initializing.")

            self._feature2memory[feature_type] = \
                FeatureMemory.construct_from_files(id_vector_file,
                                                   bg_flags_file,
                                                   feature_mat_file,
                                                   kernel_mat_file,
                                                   rw_lock)

    def remove(self, feature_type):
        """ Removes a feature type from our map, releasing its contents.

        :raise KeyError: If the given feature type does not currently map to
            anything.

        :param feature_type: The feature type to get the memory object of.
        :type feature_type: str

        """
        with self._map_lock:
            del self._feature2memory[feature_type]

    def get_feature_memory(self, feature_type):
        """ Get the underlying FeatureMemory object for a feature type.

        :raise KeyError: If the given feature type does not currently map to
            anything.

        :param feature_type: The feature type to get the memory object of.
        :type feature_type: str

        :return: FeatureMemory object associated to the given feature type.
        :rtype: FeatureMemory

        """
        with self._map_lock:
            return self._feature2memory[feature_type]

    def get_distance_kernel(self, feature_type):
        """ Get the DistanceKernel for a feature type.

        :raise KeyError: If the given feature type does not currently map to
            anything.

        :param feature_type: The feature type to get the memory object of.
        :type feature_type: str

        """
        with self._map_lock:
            return self._feature2memory[feature_type]\
                       .get_distance_kernel()

    def get_feature(self, feature_type, *clip_id_or_ids):
        """
        With respect to the given feature types, return a matrix where each row
        is the feature vector for one or more clip IDs. The given clip IDs acts
        as the index-to-clipID map for the returned matrix's rows. If repeat
        clip IDs are provided in the input, there will be repeat feature vectors
        in the returned matrix.

        :raise ValueError: If the given clip ID is not represented in the
            feature matrix.

        :param feature_type: Feature type to access:
        :type feature_type: str

        :param clip_id_or_ids: One or more integer clip IDs to retrieve the
            feature vectors for.
        :type clip_id_or_ids: int

        :return: NxM matrix, where N is the number of clip IDs requested and M
            is the length of a feature vector for this vector.
        :rtype: np.matrix

        """
        with self._map_lock:
            return self._feature2memory[feature_type]\
                       .get_feature(*clip_id_or_ids)

    def update(self, feature_type, clip_id, feature_vector, is_background=False,
               timeout=None):
        with self._map_lock:
            return self._feature2memory[feature_type]\
                       .update(clip_id, feature_vector, is_background, timeout)



"""
Example of a module that implements implementations but states that nothing is
exported by setting the helper variable to ``None``.
"""
from smqtk.tests.utils.test_plugin_get import DummyInterface


class ImplSkipModule (DummyInterface):

    @classmethod
    def is_usable(cls):
        return True

    def inst_method(self, val):
        return "skipModule"+str(val)


TEST_PLUGIN_CLASS = None



from __future__ import division, print_function
import os
import unittest
import six
from six.moves import map

from smqtk.algorithms.classifier import get_classifier_impls
from smqtk.algorithms.classifier.index_label import IndexLabelClassifier
from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement
from smqtk.tests import TEST_DATA_DIR


class TestIndexLabelClassifier (unittest.TestCase):

    EXPECTED_LABEL_VEC = list(map(six.b, [
        'label_1',
        'label_2',
        'negative',
        'label_3',
        'Kitware',
        'label_4',
    ]))

    FILEPATH_TEST_LABELS = os.path.join(TEST_DATA_DIR, 'test_labels.txt')

    def test_is_usable(self):
        # Should always be available
        self.assertTrue(IndexLabelClassifier.is_usable())

    def test_impl_findable(self):
        self.assertIn(IndexLabelClassifier.__name__,
                      get_classifier_impls())

    def test_new(self):
        c = IndexLabelClassifier(self.FILEPATH_TEST_LABELS)
        self.assertEqual(c.label_vector, self.EXPECTED_LABEL_VEC)

    def test_get_labels(self):
        c = IndexLabelClassifier(self.FILEPATH_TEST_LABELS)
        self.assertEqual(c.get_labels(), self.EXPECTED_LABEL_VEC)

    def test_configuration(self):
        cfg = IndexLabelClassifier.get_default_config()
        self.assertEqual(cfg, {'index_to_label_uri': None})

        cfg['index_to_label_uri'] = self.FILEPATH_TEST_LABELS
        c = IndexLabelClassifier.from_config(cfg)
        self.assertEqual(c.get_config(), cfg)

    def test_classify(self):
        c = IndexLabelClassifier(self.FILEPATH_TEST_LABELS)
        m_expected = {
            six.b('label_1'): 1,
            six.b('label_2'): 2,
            six.b('negative'): 3,
            six.b('label_3'): 4,
            six.b('Kitware'): 5,
            six.b('label_4'): 6,
        }

        d = DescriptorMemoryElement('test', 0)
        d.set_vector([1, 2, 3, 4, 5, 6])

        m = c._classify(d)
        self.assertEqual(m, m_expected)

    def test_classify_invalid_descriptor_dimensions(self):
        c = IndexLabelClassifier(self.FILEPATH_TEST_LABELS)
        d = DescriptorMemoryElement('test', 0)

        # One less
        d.set_vector([1, 2, 3, 4, 5])
        self.assertRaises(
            RuntimeError,
            c._classify, d
        )

        # One more
        d.set_vector([1, 2, 3, 4, 5, 6, 7])
        self.assertRaises(
            RuntimeError,
            c._classify, d
        )



from __future__ import division, print_function
import unittest

import mock

from smqtk.algorithms.descriptor_generator import get_descriptor_generator_impls
from smqtk.algorithms.descriptor_generator.colordescriptor.colordescriptor \
    import ColorDescriptor_Image_csift  # arbitrary leaf class


if ColorDescriptor_Image_csift.is_usable():

    class TestColorDescriptor (unittest.TestCase):

        def test_impl_findable(self):
            self.assertIn(ColorDescriptor_Image_csift.__name__,
                          get_descriptor_generator_impls())

        @mock.patch('smqtk.algorithms.descriptor_generator'
                    '.colordescriptor.colordescriptor'
                    '.file_utils.safe_create_dir')
        def test_configuration(self, _mock_scd):
            default_config = ColorDescriptor_Image_csift.get_default_config()
            default_config['model_directory'] = '/some/path/models/'
            default_config['work_directory'] = '/some/path/work/'

            inst = ColorDescriptor_Image_csift.from_config(default_config)
            self.assertEqual(
                default_config,
                inst.get_config()
            )
            inst2 = ColorDescriptor_Image_csift.from_config(inst.get_config())
            self.assertEqual(inst.get_config(), inst2.get_config())
            self.assertEqual(inst._model_dir, inst2._model_dir)
            self.assertEqual(inst._work_dir, inst2._work_dir)
            self.assertEqual(inst._kmeans_k, inst2._kmeans_k)
            self.assertEqual(inst._flann_target_precision,
                             inst2._flann_target_precision)
            self.assertEqual(inst._flann_sample_fraction,
                             inst2._flann_sample_fraction)
            self.assertEqual(inst._flann_autotune, inst2._flann_autotune)
            self.assertEqual(inst._use_sp, inst2._use_sp)
            self.assertEqual(inst._rand_seed, inst2._rand_seed)



import unittest

import numpy
import six
from six.moves import cPickle as pickle

from smqtk.representation.data_element.memory_element import DataMemoryElement
from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement
from smqtk.representation.descriptor_index.memory import MemoryDescriptorIndex
from smqtk.utils import merge_dict


RAND_UUID = 0


def random_descriptor():
    global RAND_UUID
    d = DescriptorMemoryElement('random', RAND_UUID)
    d.set_vector(numpy.random.rand(64))
    RAND_UUID += 1
    return d


class TestMemoryDescriptorIndex (unittest.TestCase):

    def test_is_usable(self):
        # Always usable because no dependencies.
        self.assertEqual(MemoryDescriptorIndex.is_usable(), True)

    def test_default_config(self):
        # Default should be valid for constructing a new instance.
        c = MemoryDescriptorIndex.get_default_config()
        self.assertEqual(MemoryDescriptorIndex.from_config(c).get_config(), c)

    def test_from_config_null_cache_elem(self):
        inst = MemoryDescriptorIndex.from_config({'cache_element': None})
        self.assertIsNone(inst.cache_element)
        self.assertEqual(inst._table, {})

        inst = MemoryDescriptorIndex.from_config({
            'cache_element': {
                'type': None
            }
        })
        self.assertIsNone(inst.cache_element)
        self.assertEqual(inst._table, {})

    def test_from_config_null_cache_elem_type(self):
        # An empty cache should not trigger loading on construction.
        expected_empty_cache = DataMemoryElement()
        inst = MemoryDescriptorIndex.from_config({
            'cache_element': {
                'type': 'DataMemoryElement',
                'DataMemoryElement': {'bytes': ''}
            }
        })
        self.assertEqual(inst.cache_element, expected_empty_cache)
        self.assertEqual(inst._table, {})

    def test_from_config(self):
        # Configured cache with some picked bytes
        expected_table = dict(a=1, b=2, c=3)
        expected_cache = DataMemoryElement(bytes=pickle.dumps(expected_table))
        inst = MemoryDescriptorIndex.from_config({
            'cache_element': {
                'type': 'DataMemoryElement',
                'DataMemoryElement': {'bytes': expected_cache.get_bytes()}
            }
        })
        self.assertEqual(inst.cache_element, expected_cache)
        self.assertEqual(inst._table, expected_table)

    def test_init_no_cache(self):
        inst = MemoryDescriptorIndex()
        self.assertIsNone(inst.cache_element, None)
        self.assertEqual(inst._table, {})

    def test_init_empty_cache(self):
        cache_elem = DataMemoryElement()
        inst = MemoryDescriptorIndex(cache_element=cache_elem)
        self.assertEqual(inst.cache_element, cache_elem)
        self.assertEqual(inst._table, {})

    def test_init_with_cache(self):
        d_list = (random_descriptor(), random_descriptor(),
                  random_descriptor(), random_descriptor())
        expected_table = dict((r.uuid(), r) for r in d_list)
        expected_cache = DataMemoryElement(bytes=pickle.dumps(expected_table))

        inst = MemoryDescriptorIndex(expected_cache)
        self.assertEqual(len(inst._table), 4)
        self.assertEqual(inst.cache_element, expected_cache)
        self.assertEqual(inst._table, expected_table)
        self.assertEqual(set(inst._table.values()), set(d_list))

    def test_get_config(self):
        self.assertEqual(
            MemoryDescriptorIndex().get_config(),
            MemoryDescriptorIndex.get_default_config()
        )

        self.assertEqual(
            MemoryDescriptorIndex(None).get_config(),
            MemoryDescriptorIndex.get_default_config()
        )

        empty_elem = DataMemoryElement()
        self.assertEqual(
            MemoryDescriptorIndex(empty_elem).get_config(),
            merge_dict(MemoryDescriptorIndex.get_default_config(), {
                'cache_element': {'type': 'DataMemoryElement'}
            })
        )

        dict_pickle_bytes = pickle.dumps({1: 1, 2: 2, 3: 3}, -1)
        cache_elem = DataMemoryElement(bytes=dict_pickle_bytes)
        self.assertEqual(
            MemoryDescriptorIndex(cache_elem).get_config(),
            merge_dict(MemoryDescriptorIndex.get_default_config(), {
                'cache_element': {
                    'DataMemoryElement': {
                        'bytes': dict_pickle_bytes
                    },
                    'type': 'DataMemoryElement'
                }
            })
        )

    def test_cache_table_no_cache(self):
        inst = MemoryDescriptorIndex()
        inst._table = {}
        inst.cache_table()  # should basically do nothing
        self.assertIsNone(inst.cache_element)

    def test_cache_table_empty_table(self):
        inst = MemoryDescriptorIndex(DataMemoryElement(), -1)
        inst._table = {}
        expected_table_pickle_bytes = pickle.dumps(inst._table, -1)

        inst.cache_table()
        self.assertIsNotNone(inst.cache_element)
        self.assertEqual(inst.cache_element.get_bytes(),
                         expected_table_pickle_bytes)

    def test_add_descriptor(self):
        index = MemoryDescriptorIndex()

        d1 = random_descriptor()
        index.add_descriptor(d1)
        self.assertEqual(index._table[d1.uuid()], d1)

        d2 = random_descriptor()
        index.add_descriptor(d2)
        self.assertEqual(index._table[d2.uuid()], d2)

    def test_add_many(self):
        descrs = [
            random_descriptor(),
            random_descriptor(),
            random_descriptor(),
            random_descriptor(),
            random_descriptor(),
        ]
        index = MemoryDescriptorIndex()
        index.add_many_descriptors(descrs)

        # Compare code keys of input to code keys in internal table
        self.assertEqual(set(index._table.keys()),
                         set([e.uuid() for e in descrs]))

        # Get the set of descriptors in the internal table and compare it with
        # the set of generated random descriptors.
        r_set = set()
        [r_set.add(d) for d in index._table.values()]
        self.assertEqual(
            set([e for e in descrs]),
            r_set
        )

    def test_count(self):
        index = MemoryDescriptorIndex()
        self.assertEqual(index.count(), 0)

        d1 = random_descriptor()
        index.add_descriptor(d1)
        self.assertEqual(index.count(), 1)

        d2, d3, d4 = (random_descriptor(),
                      random_descriptor(),
                      random_descriptor())
        index.add_many_descriptors([d2, d3, d4])
        self.assertEqual(index.count(), 4)

        d5 = random_descriptor()
        index.add_descriptor(d5)
        self.assertEqual(index.count(), 5)

    def test_get_descriptors(self):
        descrs = [
            random_descriptor(),   # [0]
            random_descriptor(),   # [1]
            random_descriptor(),   # [2]
            random_descriptor(),   # [3]
            random_descriptor(),   # [4]
        ]
        index = MemoryDescriptorIndex()
        index.add_many_descriptors(descrs)

        # single descriptor reference
        r = index.get_descriptor(descrs[1].uuid())
        self.assertEqual(r, descrs[1])

        # multiple descriptor reference
        r = list(index.get_many_descriptors([descrs[0].uuid(),
                                             descrs[3].uuid()]))
        self.assertEqual(len(r), 2)
        self.assertEqual(set(r),
                         {descrs[0], descrs[3]})

    def test_clear(self):
        i = MemoryDescriptorIndex()
        n = 10

        descrs = [random_descriptor() for _ in range(n)]
        i.add_many_descriptors(descrs)
        self.assertEqual(len(i), n)
        i.clear()
        self.assertEqual(len(i), 0)
        self.assertEqual(i._table, {})

    def test_has(self):
        i = MemoryDescriptorIndex()
        descrs = [random_descriptor() for _ in range(10)]
        i.add_many_descriptors(descrs)

        self.assertTrue(i.has_descriptor(descrs[4].uuid()))
        self.assertFalse(i.has_descriptor('not_an_int'))

    def test_added_descriptor_table_caching(self):
        cache_elem = DataMemoryElement(readonly=False)
        descrs = [random_descriptor() for _ in range(3)]
        expected_table = dict((r.uuid(), r) for r in descrs)

        i = MemoryDescriptorIndex(cache_elem)
        self.assertTrue(cache_elem.is_empty())

        # Should add descriptors to table, caching to writable element.
        i.add_many_descriptors(descrs)
        self.assertFalse(cache_elem.is_empty())
        self.assertEqual(pickle.loads(i.cache_element.get_bytes()),
                         expected_table)

        # Changing the internal table (remove, add) it should reflect in
        # cache
        new_d = random_descriptor()
        expected_table[new_d.uuid()] = new_d
        i.add_descriptor(new_d)
        self.assertEqual(pickle.loads(i.cache_element.get_bytes()),
                         expected_table)

        rm_d = list(expected_table.values())[0]
        del expected_table[rm_d.uuid()]
        i.remove_descriptor(rm_d.uuid())
        self.assertEqual(pickle.loads(i.cache_element.get_bytes()),
                         expected_table)

    def test_remove(self):
        i = MemoryDescriptorIndex()
        descrs = [random_descriptor() for _ in range(100)]
        i.add_many_descriptors(descrs)
        self.assertEqual(len(i), 100)
        self.assertEqual(list(i.iterdescriptors()), descrs)

        # remove singles
        i.remove_descriptor(descrs[0].uuid())
        self.assertEqual(len(i), 99)
        self.assertEqual(set(i.iterdescriptors()),
                         set(descrs[1:]))

        # remove many
        rm_d = descrs[slice(45, 80, 3)]
        i.remove_many_descriptors((d.uuid() for d in rm_d))
        self.assertEqual(len(i), 99 - len(rm_d))
        self.assertEqual(set(i.iterdescriptors()),
                         set(descrs[1:]).difference(rm_d))

    def test_iterdescrs(self):
        i = MemoryDescriptorIndex()
        descrs = [random_descriptor() for _ in range(100)]
        i.add_many_descriptors(descrs)
        self.assertEqual(set(i.iterdescriptors()),
                         set(descrs))

    def test_iterkeys(self):
        i = MemoryDescriptorIndex()
        descrs = [random_descriptor() for _ in range(100)]
        i.add_many_descriptors(descrs)
        self.assertEqual(set(i.iterkeys()),
                         set(d.uuid() for d in descrs))

    def test_iteritems(self):
        i = MemoryDescriptorIndex()
        descrs = [random_descriptor() for _ in range(100)]
        i.add_many_descriptors(descrs)
        self.assertEqual(set(six.iteritems(i)),
                         set((d.uuid(), d) for d in descrs))



import base64
import re
import six

from smqtk.exceptions import InvalidUriError, ReadOnlyError
from smqtk.representation import DataElement


class DataMemoryElement (DataElement):
    """
    In-memory representation of data stored in a byte list
    """

    # Base64 RE including URL-safe character replacements
    B64_PATTERN = '[a-zA-Z0-9+/_-]*={0,2}'
    URI_B64_RE = re.compile('^base64://(?P<base64>{})$'.format(B64_PATTERN))
    URI_DATA_B64_RE = re.compile("^data:(?P<ct>[\w/]+);base64,(?P<base64>{})$"
                                 .format(B64_PATTERN))

    @classmethod
    def is_usable(cls):
        # No dependencies
        return True

    @classmethod
    def from_uri(cls, uri):
        """
        Construct a new instance based on the given URI.

        Memory elements resolve byte-string formats. Currently, this method
        accepts a base64 using the standard and URL-safe alphabet as the python
        ``base64.urlsafe_b64decode`` module function would expect.

        This method accepts URIs in two formats:
            - ``base64://<data>``
            - ``data:<mimetype>;base64,<data>``
            - Empty string (no data)

        Filling in ``<data>`` with the actual byte string, and ``<mimetype>``
        with the actual MIMETYPE of the bytes.

        :param uri: URI string to resolve into an element instance
        :type uri: str

        :raises smqtk.exceptions.InvalidUriError: The given URI was not a base64
            format

        :return: New element instance of our type.
        :rtype: DataElement

        """
        if uri is None:
            raise InvalidUriError(uri, 'None value given')

        if len(uri) == 0:
            return DataMemoryElement('', None)

        data_b64_m = cls.URI_B64_RE.match(uri)
        if data_b64_m is not None:
            m_d = data_b64_m.groupdict()
            return DataMemoryElement.from_base64(m_d['base64'], None)

        data_b64_m = cls.URI_DATA_B64_RE.match(uri)
        if data_b64_m is not None:
            m_d = data_b64_m.groupdict()
            return DataMemoryElement.from_base64(
                m_d['base64'], m_d['ct']
            )

        raise InvalidUriError(uri, "Did not detect byte format URI")

    @classmethod
    def from_base64(cls, b64_str, content_type=None):
        """
        Create new MemoryElement instance based on a given base64 string and
        content type.

        This method accepts a base64 using the standard and URL-safe alphabet as
        the python ``base64.urlsafe_b64decode`` module function would expect.

        :param b64_str: Base64 data string.
        :type b64_str: str

        :param content_type: Content type string, or None if unknown.
        :type content_type: str | None

        :return: New MemoryElement instance containing the byte data in the
            given base64 string.
        :rtype: DataMemoryElement

        """
        if b64_str is None:
            raise ValueError("Base 64 string should not be None")
        # The decode function does not like taking unicode strings.
        # Additionally, the encoding alphabet should not container any unicode
        # symbols, so this aught to be safe.
        b64_str = str(b64_str)
        return DataMemoryElement(base64.urlsafe_b64decode(b64_str),
                                 content_type)

    # noinspection PyShadowingBuiltins
    def __init__(self, bytes=None, content_type=None, readonly=False):
        """
        Create a new DataMemoryElement from a byte string and optional content
        type.

        :param bytes: Bytes to contain. May be None to represent no bytes.
        :type bytes: None | bytes

        :param content_type: Content type of the bytes given.
        :type content_type: None | basestring

        :param readonly: If this element should allow writing or not.
        :type readonly: bool

        """
        super(DataMemoryElement, self).__init__()

        self._bytes = bytes
        self._content_type = content_type
        self._readonly = bool(readonly)

    def __repr__(self):
        return super(DataMemoryElement, self).__repr__() + \
               "{len(bytes): %d, content_type: %s, readonly: %s}" \
               % (len(self.get_bytes()), self._content_type, self._readonly)

    def get_config(self):
        return {
            "bytes": self._bytes,
            'content_type': self._content_type,
            "readonly": self._readonly,
        }

    def content_type(self):
        """
        :return: Standard type/subtype string for this data element, or None if
            the content type is unknown.
        :rtype: str or None
        """
        return self._content_type

    def is_empty(self):
        """
        Check if this element contains no bytes.

        :return: If this element contains 0 bytes.
        :rtype: bool

        """
        return not bool(self._bytes)

    def get_bytes(self):
        """
        :return: Get the byte stream for this data element.
        :rtype: bytes
        """
        return self._bytes or six.b('')

    def writable(self):
        """
        :return: if this instance supports setting bytes.
        :rtype: bool
        """
        return not self._readonly

    def set_bytes(self, b):
        """
        Set bytes to this data element in the form of a string.

        Previous content type value is maintained.

        :param b: bytes to set.
        :type b: str

        :raises ReadOnlyError: This data element can only be read from / does
            not support writing.

        """
        if not self._readonly:
            self._bytes = b
        else:
            raise ReadOnlyError("This memory element cannot be written to.")


DATA_ELEMENT_CLASS = DataMemoryElement



from __future__ import division, print_function

import collections
import json
import mock
import random
import types
import unittest

import numpy as np

from smqtk.algorithms.nn_index.lsh import LSHNearestNeighborIndex
from smqtk.algorithms.nn_index.lsh.functors import LshFunctor
from smqtk.algorithms.nn_index.lsh.functors.itq import ItqFunctor
from smqtk.algorithms.nn_index.hash_index import HashIndex
from smqtk.algorithms.nn_index.hash_index.linear import LinearHashIndex
from smqtk.algorithms.nn_index.hash_index.sklearn_balltree import \
    SkLearnBallTreeHashIndex
from smqtk.exceptions import ReadOnlyError
from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement
from smqtk.representation.descriptor_index.memory import MemoryDescriptorIndex
from smqtk.representation.key_value.memory import MemoryKeyValueStore


class DummyHashFunctor (LshFunctor):

    @classmethod
    def is_usable(cls):
        return True

    def get_config(self):
        """ stub """

    def get_hash(self, descriptor):
        """
        Dummy function that returns the bits of the integer sum of descriptor
        vector.

        :param descriptor: Descriptor vector we should generate the hash of.
        :type descriptor: np.ndarray[float]

        :return: Generated bit-vector as a numpy array of booleans.
        :rtype: np.ndarray[bool]

        """
        return np.asarray([int(c) for c in bin(int(descriptor.sum()))[2:]],
                             bool)


class TestLshIndex (unittest.TestCase):

    def test_is_usable(self):
        # Should always be usable since this is a shell class.
        self.assertTrue(LSHNearestNeighborIndex.is_usable())

    def test_configuration(self):
        c = LSHNearestNeighborIndex.get_default_config()

        # Check that default is in JSON format and is decoded to the same
        # result.
        self.assertEqual(json.loads(json.dumps(c)), c)

        # Make a simple configuration
        # - ItqFunctor should always be available since it has no dependencies.
        c['lsh_functor']['type'] = 'ItqFunctor'
        c['descriptor_index']['type'] = 'MemoryDescriptorIndex'
        c['hash2uuids_kvstore']['type'] = 'MemoryKeyValueStore'
        c['hash_index']['type'] = 'LinearHashIndex'
        index = LSHNearestNeighborIndex.from_config(c)

        self.assertIsInstance(index.lsh_functor, ItqFunctor)
        self.assertIsInstance(index.descriptor_index, MemoryDescriptorIndex)
        self.assertIsInstance(index.hash_index, LinearHashIndex)
        self.assertIsInstance(index.hash2uuids_kvstore, MemoryKeyValueStore)

        # Can convert instance config to JSON
        self.assertEqual(
            json.loads(json.dumps(index.get_config())),
            index.get_config()
        )

    def test_configuration_none_HI(self):
        c = LSHNearestNeighborIndex.get_default_config()

        # Check that default is in JSON format and is decoded to the same
        # result.
        self.assertEqual(json.loads(json.dumps(c)), c)

        # Make a simple configuration
        c['lsh_functor']['type'] = 'ItqFunctor'
        c['descriptor_index']['type'] = 'MemoryDescriptorIndex'
        c['hash2uuids_kvstore']['type'] = 'MemoryKeyValueStore'
        c['hash_index']['type'] = None
        index = LSHNearestNeighborIndex.from_config(c)

        self.assertIsInstance(index.lsh_functor, ItqFunctor)
        self.assertIsInstance(index.descriptor_index, MemoryDescriptorIndex)
        self.assertIsNone(index.hash_index)
        self.assertIsInstance(index.hash2uuids_kvstore, MemoryKeyValueStore)

        # Can convert instance config to JSON
        self.assertEqual(
            json.loads(json.dumps(index.get_config())),
            index.get_config()
        )

    def test_get_dist_func_euclidean(self):
        f = LSHNearestNeighborIndex._get_dist_func('euclidean')
        self.assertIsInstance(f, types.FunctionType)
        self.assertAlmostEqual(
            f(np.array([0, 0]), np.array([0, 1])),
            1.0
        )

    def test_get_dist_func_cosine(self):
        f = LSHNearestNeighborIndex._get_dist_func('cosine')
        self.assertIsInstance(f, types.FunctionType)
        self.assertAlmostEqual(
            f(np.array([1, 0]), np.array([0, 1])),
            1.0
        )
        self.assertAlmostEqual(
            f(np.array([1, 0]), np.array([1, 1])),
            0.5
        )

    def test_get_dist_func_hik(self):
        f = LSHNearestNeighborIndex._get_dist_func('hik')
        self.assertIsInstance(f, types.FunctionType)
        self.assertAlmostEqual(
            f(np.array([0, 0]), np.array([0, 1])),
            1.0
        )
        self.assertAlmostEqual(
            f(np.array([1, 0]), np.array([0, 1])),
            1.0
        )
        self.assertAlmostEqual(
            f(np.array([1, 1]), np.array([0, 1])),
            0.0
        )

    def test_get_dist_func_invalid_string(self):
        self.assertRaises(
            ValueError,
            LSHNearestNeighborIndex._get_dist_func,
            'not-valid-string'
        )

    def test_count_empty_hash2uid(self):
        """
        Test that an empty hash-to-uid mapping results in a 0 return regardless
        of descriptor-set state.
        """
        descr_set = MemoryDescriptorIndex()
        hash_kvs = MemoryKeyValueStore()
        self.assertEqual(descr_set.count(), 0)
        self.assertEqual(hash_kvs.count(), 0)

        lsh = LSHNearestNeighborIndex(DummyHashFunctor(), descr_set, hash_kvs)
        self.assertEqual(lsh.count(), 0)

        # Additions to the descriptor-set should not impact LSH index "size"
        lsh.descriptor_index.add_descriptor(DescriptorMemoryElement('t', 0))
        self.assertEqual(lsh.descriptor_index.count(), 1)
        self.assertEqual(lsh.hash2uuids_kvstore.count(), 0)
        self.assertEqual(lsh.count(), 0)

        lsh.descriptor_index.add_descriptor(DescriptorMemoryElement('t', 1))
        self.assertEqual(lsh.descriptor_index.count(), 2)
        self.assertEqual(lsh.hash2uuids_kvstore.count(), 0)
        self.assertEqual(lsh.count(), 0)

        lsh.hash2uuids_kvstore.add(0, {0})
        self.assertEqual(lsh.descriptor_index.count(), 2)
        self.assertEqual(lsh.count(), 1)

        lsh.hash2uuids_kvstore.add(0, {0, 1})
        self.assertEqual(lsh.descriptor_index.count(), 2)
        self.assertEqual(lsh.count(), 2)

        lsh.hash2uuids_kvstore.add(0, {0, 1, 2})
        self.assertEqual(lsh.descriptor_index.count(), 2)
        self.assertEqual(lsh.count(), 3)

    def test_build_index_read_only(self):
        index = LSHNearestNeighborIndex(DummyHashFunctor(),
                                        MemoryDescriptorIndex(),
                                        MemoryKeyValueStore(), read_only=True)
        self.assertRaises(
            ReadOnlyError,
            index._build_index, []
        )

    def test_build_index_fresh_build(self):
        descr_index = MemoryDescriptorIndex()
        hash_kvs = MemoryKeyValueStore()
        index = LSHNearestNeighborIndex(DummyHashFunctor(),
                                        descr_index, hash_kvs)

        descriptors = [
            DescriptorMemoryElement('t', 0),
            DescriptorMemoryElement('t', 1),
            DescriptorMemoryElement('t', 2),
            DescriptorMemoryElement('t', 3),
            DescriptorMemoryElement('t', 4),
        ]
        # Vectors of length 1 for easy dummy hashing prediction.
        for i, d in enumerate(descriptors):
            d.set_vector(np.ones(1, float) * i)
        index.build_index(descriptors)

        # Make sure descriptors are now in attached index and in key-value-store
        self.assertEqual(descr_index.count(), 5)
        for d in descriptors:
            self.assertIn(d, descr_index)
        # Dummy hash function bins sum of descriptor vectors.
        self.assertEqual(hash_kvs.count(), 5)
        for i in range(5):
            self.assertSetEqual(hash_kvs.get(i), {i})

    def test_build_index_fresh_build_with_hash_index(self):
        descr_index = MemoryDescriptorIndex()
        hash_kvs = MemoryKeyValueStore()
        linear_hi = LinearHashIndex()  # simplest hash index, heap-sorts.
        index = LSHNearestNeighborIndex(DummyHashFunctor(),
                                        descr_index, hash_kvs, linear_hi)

        descriptors = [
            DescriptorMemoryElement('t', 0),
            DescriptorMemoryElement('t', 1),
            DescriptorMemoryElement('t', 2),
            DescriptorMemoryElement('t', 3),
            DescriptorMemoryElement('t', 4),
        ]
        # Vectors of length 1 for easy dummy hashing prediction.
        for i, d in enumerate(descriptors):
            d.set_vector(np.ones(1, float) * i)
        index.build_index(descriptors)
        # Hash index should have been built with hash vectors, and linearHI
        # converts those to integers for storage.
        self.assertEqual(linear_hi.index, {0, 1, 2, 3, 4})

    def test_update_index_read_only(self):
        index = LSHNearestNeighborIndex(DummyHashFunctor(),
                                        MemoryDescriptorIndex(),
                                        MemoryKeyValueStore(), read_only=True)
        self.assertRaises(
            ReadOnlyError,
            index._update_index, []
        )

    def test_update_index_no_existing_index(self):
        # Test that calling update_index with no existing index acts like
        # building the index fresh.  This test is basically the same as
        # test_build_index_fresh_build but using update_index instead.
        descr_index = MemoryDescriptorIndex()
        hash_kvs = MemoryKeyValueStore()
        index = LSHNearestNeighborIndex(DummyHashFunctor(),
                                        descr_index, hash_kvs)

        descriptors = [
            DescriptorMemoryElement('t', 0),
            DescriptorMemoryElement('t', 1),
            DescriptorMemoryElement('t', 2),
            DescriptorMemoryElement('t', 3),
            DescriptorMemoryElement('t', 4),
        ]
        # Vectors of length 1 for easy dummy hashing prediction.
        for d in descriptors:
            d.set_vector(np.ones(1, float) * d.uuid())
        index.update_index(descriptors)

        # Make sure descriptors are now in attached index and in key-value-store
        self.assertEqual(descr_index.count(), 5)
        for d in descriptors:
            self.assertIn(d, descr_index)
        # Dummy hash function bins sum of descriptor vectors.
        self.assertEqual(hash_kvs.count(), 5)
        for i in range(5):
            self.assertSetEqual(hash_kvs.get(i), {i})

    def test_update_index_add_new_descriptors(self):
        # Test that calling update index after a build index causes index
        # components to be properly updated.
        descr_index = MemoryDescriptorIndex()
        hash_kvs = MemoryKeyValueStore()
        index = LSHNearestNeighborIndex(DummyHashFunctor(),
                                        descr_index, hash_kvs)
        descriptors1 = [
            DescriptorMemoryElement('t', 0),
            DescriptorMemoryElement('t', 1),
            DescriptorMemoryElement('t', 2),
            DescriptorMemoryElement('t', 3),
            DescriptorMemoryElement('t', 4),
        ]
        descriptors2 = [
            DescriptorMemoryElement('t', 5),
            DescriptorMemoryElement('t', 6),
        ]
        # Vectors of length 1 for easy dummy hashing prediction.
        for d in descriptors1 + descriptors2:
            d.set_vector(np.ones(1, float) * d.uuid())

        # Build initial index.
        index.build_index(descriptors1)
        self.assertEqual(descr_index.count(), 5)
        for d in descriptors1:
            self.assertIn(d, descr_index)
        for d in descriptors2:
            self.assertNotIn(d, descr_index)
        # Dummy hash function bins sum of descriptor vectors.
        self.assertEqual(hash_kvs.count(), 5)
        for i in range(5):
            self.assertSetEqual(hash_kvs.get(i), {i})

        # Update index and check that components have new data.
        index.update_index(descriptors2)
        self.assertEqual(descr_index.count(), 7)
        for d in descriptors1 + descriptors2:
            self.assertIn(d, descr_index)
        # Dummy hash function bins sum of descriptor vectors.
        self.assertEqual(hash_kvs.count(), 7)
        for i in range(7):
            self.assertSetEqual(hash_kvs.get(i), {i})

    def test_update_index_duplicate_descriptors(self):
        """
        Test that updating a built index with the same descriptors results in
        idempotent behavior.
        """
        descr_index = MemoryDescriptorIndex()
        hash_kvs = MemoryKeyValueStore()
        index = LSHNearestNeighborIndex(DummyHashFunctor(),
                                        descr_index, hash_kvs)

        # Identical Descriptors to build and update on (different instances)
        descriptors1 = [
            DescriptorMemoryElement('t', 0).set_vector([0]),
            DescriptorMemoryElement('t', 1).set_vector([1]),
            DescriptorMemoryElement('t', 2).set_vector([2]),
            DescriptorMemoryElement('t', 3).set_vector([3]),
            DescriptorMemoryElement('t', 4).set_vector([4]),
        ]
        descriptors2 = [
            DescriptorMemoryElement('t', 0).set_vector([0]),
            DescriptorMemoryElement('t', 1).set_vector([1]),
            DescriptorMemoryElement('t', 2).set_vector([2]),
            DescriptorMemoryElement('t', 3).set_vector([3]),
            DescriptorMemoryElement('t', 4).set_vector([4]),
        ]

        index.build_index(descriptors1)
        index.update_index(descriptors2)

        assert descr_index.count() == 5
        # Above descriptors should be considered "in" the descriptor set now.
        for d in descriptors1:
            assert d in descr_index
        for d in descriptors2:
            assert d in descr_index
        # Known hashes of the above descriptors should be in the KVS
        assert set(hash_kvs.keys()) == {0, 1, 2, 3, 4}
        assert hash_kvs.get(0) == {0}
        assert hash_kvs.get(1) == {1}
        assert hash_kvs.get(2) == {2}
        assert hash_kvs.get(3) == {3}
        assert hash_kvs.get(4) == {4}

    def test_update_index_similar_descriptors(self):
        """
        Test that updating a built index with similar descriptors (same
        vectors, different UUIDs) results in contained structures having an
        expected state.
        """
        descr_index = MemoryDescriptorIndex()
        hash_kvs = MemoryKeyValueStore()
        index = LSHNearestNeighborIndex(DummyHashFunctor(),
                                        descr_index, hash_kvs)

        # Similar Descriptors to build and update on (different instances)
        descriptors1 = [
            DescriptorMemoryElement('t', 0).set_vector([0]),
            DescriptorMemoryElement('t', 1).set_vector([1]),
            DescriptorMemoryElement('t', 2).set_vector([2]),
            DescriptorMemoryElement('t', 3).set_vector([3]),
            DescriptorMemoryElement('t', 4).set_vector([4]),
        ]
        descriptors2 = [
            DescriptorMemoryElement('t', 5).set_vector([0]),
            DescriptorMemoryElement('t', 6).set_vector([1]),
            DescriptorMemoryElement('t', 7).set_vector([2]),
            DescriptorMemoryElement('t', 8).set_vector([3]),
            DescriptorMemoryElement('t', 9).set_vector([4]),
        ]

        index.build_index(descriptors1)
        index.update_index(descriptors2)

        assert descr_index.count() == 10
        # Above descriptors should be considered "in" the descriptor set now.
        for d in descriptors1:
            assert d in descr_index
        for d in descriptors2:
            assert d in descr_index
        # Known hashes of the above descriptors should be in the KVS
        assert set(hash_kvs.keys()) == {0, 1, 2, 3, 4}
        assert hash_kvs.get(0) == {0, 5}
        assert hash_kvs.get(1) == {1, 6}
        assert hash_kvs.get(2) == {2, 7}
        assert hash_kvs.get(3) == {3, 8}
        assert hash_kvs.get(4) == {4, 9}

    def test_update_index_existing_descriptors_frozenset(self):
        """
        Same as ``test_update_index_similar_descriptors`` but testing that
        we can update the index when seeded with structures with existing
        values.
        """
        # Similar Descriptors to build and update on (different instances)
        descriptors1 = [
            DescriptorMemoryElement('t', 0).set_vector([0]),
            DescriptorMemoryElement('t', 1).set_vector([1]),
            DescriptorMemoryElement('t', 2).set_vector([2]),
            DescriptorMemoryElement('t', 3).set_vector([3]),
            DescriptorMemoryElement('t', 4).set_vector([4]),
        ]
        descriptors2 = [
            DescriptorMemoryElement('t', 5).set_vector([0]),
            DescriptorMemoryElement('t', 6).set_vector([1]),
            DescriptorMemoryElement('t', 7).set_vector([2]),
            DescriptorMemoryElement('t', 8).set_vector([3]),
            DescriptorMemoryElement('t', 9).set_vector([4]),
        ]

        descr_index = MemoryDescriptorIndex()
        descr_index.add_many_descriptors(descriptors1)

        hash_kvs = MemoryKeyValueStore()
        hash_kvs.add(0, frozenset({0}))
        hash_kvs.add(1, frozenset({1}))
        hash_kvs.add(2, frozenset({2}))
        hash_kvs.add(3, frozenset({3}))
        hash_kvs.add(4, frozenset({4}))

        index = LSHNearestNeighborIndex(DummyHashFunctor(),
                                        descr_index, hash_kvs)
        index.update_index(descriptors2)

        assert descr_index.count() == 10
        # Above descriptors should be considered "in" the descriptor set now.
        for d in descriptors1:
            assert d in descr_index
        for d in descriptors2:
            assert d in descr_index
        # Known hashes of the above descriptors should be in the KVS
        assert set(hash_kvs.keys()) == {0, 1, 2, 3, 4}
        assert hash_kvs.get(0) == {0, 5}
        assert hash_kvs.get(1) == {1, 6}
        assert hash_kvs.get(2) == {2, 7}
        assert hash_kvs.get(3) == {3, 8}
        assert hash_kvs.get(4) == {4, 9}

    def test_update_index_with_hash_index(self):
        # Similar test to `test_update_index_add_new_descriptors` but with a
        # linear hash index.
        descr_index = MemoryDescriptorIndex()
        hash_kvs = MemoryKeyValueStore()
        linear_hi = LinearHashIndex()  # simplest hash index, heap-sorts.
        index = LSHNearestNeighborIndex(DummyHashFunctor(),
                                        descr_index, hash_kvs, linear_hi)

        descriptors1 = [
            DescriptorMemoryElement('t', 0),
            DescriptorMemoryElement('t', 1),
            DescriptorMemoryElement('t', 2),
            DescriptorMemoryElement('t', 3),
            DescriptorMemoryElement('t', 4),
        ]
        descriptors2 = [
            DescriptorMemoryElement('t', 5),
            DescriptorMemoryElement('t', 6),
        ]
        # Vectors of length 1 for easy dummy hashing prediction.
        for d in descriptors1 + descriptors2:
            d.set_vector(np.ones(1, float) * d.uuid())

        # Build initial index.
        index.build_index(descriptors1)
        # Initial hash index should only encode hashes for first batch of
        # descriptors.
        self.assertSetEqual(linear_hi.index, {0, 1, 2, 3, 4})

        # Update index and check that components have new data.
        index.update_index(descriptors2)
        # Now the hash index should include all descriptor hashes.
        self.assertSetEqual(linear_hi.index, {0, 1, 2, 3, 4, 5, 6})

    def test_remove_from_index_read_only(self):
        d_set = MemoryDescriptorIndex()
        hash_kvs = MemoryKeyValueStore()
        idx = LSHNearestNeighborIndex(DummyHashFunctor(), d_set, hash_kvs,
                                      read_only=True)
        self.assertRaises(
            ReadOnlyError,
            idx.remove_from_index,
            ['uid1', 'uid2']
        )

    def test_remove_from_index_no_existing_index(self):
        # Test that attempting to remove from an instance with no existing
        # index (meaning empty descriptor-set and key-value-store) results in
        # a key error.
        d_set = MemoryDescriptorIndex()
        hash_kvs = MemoryKeyValueStore()
        idx = LSHNearestNeighborIndex(DummyHashFunctor(), d_set, hash_kvs)
        self.assertRaisesRegexp(
            KeyError,
            'uid1',
            idx.remove_from_index,
            ['uid1']
        )

    def test_remove_from_index_invalid_uid(self):
        # Test that attempting to remove a single invalid UID causes a key
        # error and does not affect index.

        # Descriptors are 1 dim, value == index.
        descriptors = [
            DescriptorMemoryElement('t', 0),
            DescriptorMemoryElement('t', 1),
            DescriptorMemoryElement('t', 2),
            DescriptorMemoryElement('t', 3),
            DescriptorMemoryElement('t', 4),
        ]
        # Vectors of length 1 for easy dummy hashing prediction.
        for d in descriptors:
            d.set_vector(np.ones(1, float) * d.uuid())
        # uid -> descriptor
        expected_dset_table = {
            0: descriptors[0],
            1: descriptors[1],
            2: descriptors[2],
            3: descriptors[3],
            4: descriptors[4],
        }
        # hash int -> set[uid]
        expected_kvs_table = {
            0: {0},
            1: {1},
            2: {2},
            3: {3},
            4: {4},
        }

        d_set = MemoryDescriptorIndex()
        hash_kvs = MemoryKeyValueStore()
        idx = LSHNearestNeighborIndex(DummyHashFunctor(), d_set, hash_kvs)
        idx.build_index(descriptors)
        # Assert we have the correct expected values
        self.assertEqual(idx.descriptor_index._table, expected_dset_table)
        self.assertEqual(idx.hash2uuids_kvstore._table, expected_kvs_table)

        # Attempt to remove descriptor with a UID we did not build with.
        self.assertRaisesRegexp(
            KeyError, '5',
            idx.remove_from_index, [5]
        )
        # Index should not have been modified.
        self.assertEqual(idx.descriptor_index._table, expected_dset_table)
        self.assertEqual(idx.hash2uuids_kvstore._table, expected_kvs_table)

        # Attempt to remove multiple UIDs, one valid and one invalid
        self.assertRaisesRegexp(
            KeyError, '5',
            idx.remove_from_index, [2, 5]
        )
        # Index should not have been modified.
        self.assertEqual(idx.descriptor_index._table, expected_dset_table)
        self.assertEqual(idx.hash2uuids_kvstore._table, expected_kvs_table)

    def test_remove_from_index(self):
        # Test that removing by UIDs does the correct thing.

        # Descriptors are 1 dim, value == index.
        descriptors = [
            DescriptorMemoryElement('t', 0),
            DescriptorMemoryElement('t', 1),
            DescriptorMemoryElement('t', 2),
            DescriptorMemoryElement('t', 3),
            DescriptorMemoryElement('t', 4),
        ]
        # Vectors of length 1 for easy dummy hashing prediction.
        for d in descriptors:
            d.set_vector(np.ones(1, float) * d.uuid())
        d_set = MemoryDescriptorIndex()
        hash_kvs = MemoryKeyValueStore()
        idx = LSHNearestNeighborIndex(DummyHashFunctor(), d_set, hash_kvs)
        idx.build_index(descriptors)

        # Attempt removing 1 uid.
        idx.remove_from_index([3])
        self.assertEqual(idx.descriptor_index._table, {
            0: descriptors[0],
            1: descriptors[1],
            2: descriptors[2],
            4: descriptors[4],
        })
        self.assertEqual(idx.hash2uuids_kvstore._table, {
            0: {0},
            1: {1},
            2: {2},
            4: {4},
        })

    def test_remove_from_index_shared_hashes(self):
        """
        Test that removing a descriptor (by UID) that shares a hash with other
        descriptors does not trigger removal of its hash.
        """
        # Simulate descriptors all hashing to the same hash value: 0
        hash_func = DummyHashFunctor()
        hash_func.get_hash = mock.Mock(return_value=np.asarray([0], bool))

        d_set = MemoryDescriptorIndex()
        hash2uids_kvs = MemoryKeyValueStore()
        idx = LSHNearestNeighborIndex(hash_func, d_set, hash2uids_kvs)

        # Descriptors are 1 dim, value == index.
        descriptors = [
            DescriptorMemoryElement('t', 0),
            DescriptorMemoryElement('t', 1),
            DescriptorMemoryElement('t', 2),
            DescriptorMemoryElement('t', 3),
            DescriptorMemoryElement('t', 4),
        ]
        # Vectors of length 1 for easy dummy hashing prediction.
        for d in descriptors:
            d.set_vector(np.ones(1, float) * d.uuid())
        idx.build_index(descriptors)
        # We expect the descriptor-set and kvs to look like the following now:
        self.assertDictEqual(d_set._table, {
            0: descriptors[0],
            1: descriptors[1],
            2: descriptors[2],
            3: descriptors[3],
            4: descriptors[4],
        })
        self.assertDictEqual(hash2uids_kvs._table, {0: {0, 1, 2, 3, 4}})

        # Mock out hash index as if we had an implementation so we can check
        # call to its remove_from_index method.
        idx.hash_index = mock.Mock(spec=HashIndex)

        idx.remove_from_index([2, 4])

        # Only uid 2 and 4 descriptors should be gone from d-set, kvs should
        # still have the 0 key and its set value should only contain uids 0, 1
        # and 3.  `hash_index.remove_from_index` should not be called because
        # no hashes should be marked for removal.
        self.assertDictEqual(d_set._table, {
            0: descriptors[0],
            1: descriptors[1],
            3: descriptors[3],
        })
        self.assertDictEqual(hash2uids_kvs._table, {0: {0, 1, 3}})
        idx.hash_index.remove_from_index.assert_not_called()

    def test_remove_from_index_shared_hashes_partial(self):
        """
        Test that only some hashes are removed from the hash index, but not
        others when those hashes still refer to other descriptors.
        """
        # Simulate initial state with some descriptor hashed to one value and
        # other descriptors hashed to another.

        # Vectors of length 1 for easy dummy hashing prediction.
        descriptors = [
            DescriptorMemoryElement('t', 0).set_vector([0]),
            DescriptorMemoryElement('t', 1).set_vector([1]),
            DescriptorMemoryElement('t', 2).set_vector([2]),
            DescriptorMemoryElement('t', 3).set_vector([3]),
            DescriptorMemoryElement('t', 4).set_vector([4]),
        ]

        # Dummy hash function to do the simulated thing
        hash_func = DummyHashFunctor()
        hash_func.get_hash = mock.Mock(
            # Vectors of even sum hash to 0, odd to 1.
            side_effect=lambda vec: [vec.sum() % 2]
        )

        d_set = MemoryDescriptorIndex()
        d_set._table = {
            0: descriptors[0],
            1: descriptors[1],
            2: descriptors[2],
            3: descriptors[3],
            4: descriptors[4],
        }

        hash2uid_kvs = MemoryKeyValueStore()
        hash2uid_kvs._table = {
            0: {0, 2, 4},
            1: {1, 3},
        }

        idx = LSHNearestNeighborIndex(hash_func, d_set, hash2uid_kvs)
        idx.hash_index = mock.Mock(spec=HashIndex)

        idx.remove_from_index([1, 2, 3])
        # Check that only one hash vector was passed to hash_index's removal
        # method (deque of hash-code vectors).
        idx.hash_index.remove_from_index.assert_called_once_with(
            collections.deque([
                [1],
            ])
        )
        self.assertDictEqual(d_set._table, {
            0: descriptors[0],
            4: descriptors[4],
        })
        self.assertDictEqual(hash2uid_kvs._table, {0: {0, 4}})


class TestLshIndexAlgorithms (unittest.TestCase):
    """
    Various tests on the ``nn`` method for different inputs and parameters.
    """

    RANDOM_SEED = 0

    def _make_ftor_itq(self, bits=32):
        itq_ftor = ItqFunctor(bit_length=bits, random_seed=self.RANDOM_SEED)

        def itq_fit(D):
            itq_ftor.fit(D)

        return itq_ftor, itq_fit

    # noinspection PyMethodMayBeStatic
    def _make_hi_linear(self):
        return LinearHashIndex()

    def _make_hi_balltree(self):
        return SkLearnBallTreeHashIndex(random_seed=self.RANDOM_SEED)

    #
    # Test LSH with random vectors
    #
    def _random_euclidean(self, hash_ftor, hash_idx,
                          ftor_train_hook=lambda d: None):
        # :param hash_ftor: Hash function class for generating hash codes for
        #   descriptors.
        # :param hash_idx: Hash index instance to use in local LSH algo
        #   instance.
        # :param ftor_train_hook: Function for training functor if necessary.

        # make random descriptors
        i = 1000
        dim = 256
        td = []
        np.random.seed(self.RANDOM_SEED)
        for j in range(i):
            d = DescriptorMemoryElement('random', j)
            d.set_vector(np.random.rand(dim))
            td.append(d)

        ftor_train_hook(td)

        di = MemoryDescriptorIndex()
        kvstore = MemoryKeyValueStore()
        index = LSHNearestNeighborIndex(hash_ftor, di, kvstore,
                                        hash_index=hash_idx,
                                        distance_method='euclidean')
        index.build_index(td)

        # test query from build set -- should return same descriptor when k=1
        q = td[255]
        r, dists = index.nn(q, 1)
        self.assertEqual(r[0], q)

        # test query very near a build vector
        td_q = td[0]
        q = DescriptorMemoryElement('query', i)
        v = td_q.vector().copy()
        v_min = max(v.min(), 0.1)
        v[0] += v_min
        v[dim-1] -= v_min
        q.set_vector(v)
        r, dists = index.nn(q, 1)
        self.assertFalse(np.array_equal(q.vector(), td_q.vector()))
        self.assertEqual(r[0], td_q)

        # random query
        q = DescriptorMemoryElement('query', i+1)
        q.set_vector(np.random.rand(dim))

        # for any query of size k, results should at least be in distance order
        r, dists = index.nn(q, 10)
        for j in range(1, len(dists)):
            self.assertGreater(dists[j], dists[j-1])
        r, dists = index.nn(q, i)
        for j in range(1, len(dists)):
            self.assertGreater(dists[j], dists[j-1])

    def test_random_euclidean__itq__None(self):
        ftor, fit = self._make_ftor_itq()
        self._random_euclidean(ftor, None, fit)

    def test_random_euclidean__itq__linear(self):
        ftor, fit = self._make_ftor_itq()
        hi = self._make_hi_linear()
        self._random_euclidean(ftor, hi, fit)

    def test_random_euclidean__itq__balltree(self):
        ftor, fit = self._make_ftor_itq()
        hi = self._make_hi_balltree()
        self._random_euclidean(ftor, hi, fit)

    #
    # Test unit vectors
    #
    def _known_unit(self, hash_ftor, hash_idx, dist_method,
                    ftor_train_hook=lambda d: None):
        ###
        # Unit vectors - Equal distance
        #
        dim = 5
        test_descriptors = []
        for i in range(dim):
            v = np.zeros(dim, float)
            v[i] = 1.
            d = DescriptorMemoryElement('unit', i)
            d.set_vector(v)
            test_descriptors.append(d)

        ftor_train_hook(test_descriptors)

        di = MemoryDescriptorIndex()
        kvstore = MemoryKeyValueStore()
        index = LSHNearestNeighborIndex(hash_ftor, di, kvstore,
                                        hash_index=hash_idx,
                                        distance_method=dist_method)
        index.build_index(test_descriptors)

        # query with zero vector
        # -> all modeled descriptors have no intersection, dists should be 1.0,
        #    or maximum distance by histogram intersection
        q = DescriptorMemoryElement('query', 0)
        q.set_vector(np.zeros(dim, float))
        r, dists = index.nn(q, dim)
        # All dists should be 1.0, r order doesn't matter
        for d in dists:
            self.assertEqual(d, 1.)

        # query with index element
        q = test_descriptors[3]
        r, dists = index.nn(q, 1)
        self.assertEqual(r[0], q)
        self.assertEqual(dists[0], 0.)

        r, dists = index.nn(q, dim)
        self.assertEqual(r[0], q)
        self.assertEqual(dists[0], 0.)

    def test_known_unit__euclidean__itq__None(self):
        ftor, fit = self._make_ftor_itq(5)
        self._known_unit(ftor, None, 'euclidean', fit)

    def test_known_unit__hik__itq__None(self):
        ftor, fit = self._make_ftor_itq(5)
        self._known_unit(ftor, None, 'hik', fit)

    def test_known_unit__euclidean__itq__linear(self):
        ftor, fit = self._make_ftor_itq(5)
        hi = self._make_hi_linear()
        self._known_unit(ftor, hi, 'euclidean', fit)

    def test_known_unit__hik__itq__linear(self):
        ftor, fit = self._make_ftor_itq(5)
        hi = self._make_hi_linear()
        self._known_unit(ftor, hi, 'hik', fit)

    def test_known_unit__euclidean__itq__balltree(self):
        ftor, fit = self._make_ftor_itq(5)
        hi = self._make_hi_balltree()
        self._known_unit(ftor, hi, 'euclidean', fit)

    def test_known_unit__hik__itq__balltree(self):
        ftor, fit = self._make_ftor_itq(5)
        hi = self._make_hi_balltree()
        self._known_unit(ftor, hi, 'hik', fit)

    #
    # Test with known vectors and euclidean dist
    #
    def _known_ordered_euclidean(self, hash_ftor, hash_idx,
                                 ftor_train_hook=lambda d: None):
        # make vectors to return in a known euclidean distance order
        i = 1000
        test_descriptors = []
        for j in range(i):
            d = DescriptorMemoryElement('ordered', j)
            d.set_vector(np.array([j, j*2], float))
            test_descriptors.append(d)
        random.shuffle(test_descriptors)

        ftor_train_hook(test_descriptors)

        di = MemoryDescriptorIndex()
        kvstore = MemoryKeyValueStore()
        index = LSHNearestNeighborIndex(hash_ftor, di, kvstore,
                                        hash_index=hash_idx,
                                        distance_method='euclidean')
        index.build_index(test_descriptors)

        # Since descriptors were built in increasing distance from (0,0),
        # returned descriptors for a query of [0,0] should be in index order.
        q = DescriptorMemoryElement('query', i)
        q.set_vector(np.array([0, 0], float))
        # top result should have UUID == 0 (nearest to query)
        r, dists = index.nn(q, 5)
        self.assertEqual(r[0].uuid(), 0)
        self.assertEqual(r[1].uuid(), 1)
        self.assertEqual(r[2].uuid(), 2)
        self.assertEqual(r[3].uuid(), 3)
        self.assertEqual(r[4].uuid(), 4)
        # global search should be in complete order
        r, dists = index.nn(q, i)
        for j, d, dist in zip(range(i), r, dists):
            self.assertEqual(d.uuid(), j)

    def test_known_ordered_euclidean__itq__None(self):
        ftor, fit = self._make_ftor_itq(1)
        self._known_ordered_euclidean(ftor, None, fit)

    def test_known_ordered_euclidean__itq__linear(self):
        ftor, fit = self._make_ftor_itq(1)
        hi = self._make_hi_linear()
        self._known_ordered_euclidean(ftor, hi, fit)

    def test_known_ordered_euclidean__itq__balltree(self):
        ftor, fit = self._make_ftor_itq(1)
        hi = self._make_hi_balltree()
        self._known_ordered_euclidean(ftor, hi, fit)



"""
Example where classes are found in module by introspecting parent classes.
"""
from smqtk.tests.utils.test_plugin_get import DummyInterface


class ImplFoo (DummyInterface):

    @classmethod
    def is_usable(cls):
        return True

    def inst_method(self, val):
        return 'foo'+str(val)


class ImplBar (DummyInterface):

    @classmethod
    def is_usable(cls):
        return True

    def inst_method(self, val):
        return 'bar'+str(val)


class ImplNotUsable (DummyInterface):

    @classmethod
    def is_usable(cls):
        return False

    def inst_method(self, val):
        return 'notUsable'+str(val)


class SomethingElse (list):
    pass



import mimetypes
import os

import flask
import requests

from smqtk.algorithms import (
    get_descriptor_generator_impls,
    get_nn_index_impls
)
from smqtk.representation import (
    DescriptorElementFactory,
    get_descriptor_index_impls,
)
from smqtk.representation.data_element.file_element import DataFileElement
from smqtk.representation.data_element.memory_element import DataMemoryElement
from smqtk.representation.data_element.url_element import DataUrlElement
from smqtk.utils import plugin
from smqtk.utils import merge_dict
from smqtk.web import SmqtkWebApp

from six.moves import map

MIMETYPES = mimetypes.MimeTypes()


SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))


class NearestNeighborServiceServer (SmqtkWebApp):
    """
    Simple server that takes in a specification of the following form:

        /nn/<path:uri>[?...]

    Computes the nearest neighbor index for the given data and returns a list
    of nearest neighbors in the following format

    Standard return JSON::
    {
        "success": <bool>,
        "neighbors": [ <float>, ... ]
        "message": <string>,
        "reference_uri": <uri>
    }
    """

    @classmethod
    def is_usable(cls):
        return True

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        c = super(NearestNeighborServiceServer, cls).get_default_config()
        merge_dict(c, {
            "descriptor_factory": DescriptorElementFactory.get_default_config(),
            "descriptor_generator":
                plugin.make_config(get_descriptor_generator_impls()),
            "nn_index": plugin.make_config(get_nn_index_impls()),
            "descriptor_index":
                plugin.make_config(get_descriptor_index_impls()),
            "update_descriptor_index": False,
        })
        return c

    def __init__(self, json_config):
        """
        Initialize application based of supplied JSON configuration

        :param json_config: JSON configuration dictionary
        :type json_config: dict

        """
        super(NearestNeighborServiceServer, self).__init__(json_config)

        self.update_index = json_config['update_descriptor_index']

        # Descriptor factory setup
        self._log.info("Initializing DescriptorElementFactory")
        self.descr_elem_factory = DescriptorElementFactory.from_config(
            self.json_config['descriptor_factory']
        )

        #: :type: smqtk.representation.DescriptorIndex | None
        self.descr_index = None
        if self.update_index:
            self._log.info("Initializing DescriptorIndex to update")
            #: :type: smqtk.representation.DescriptorIndex | None
            self.descr_index = plugin.from_plugin_config(
                json_config['descriptor_index'],
                get_descriptor_index_impls()
            )

        #: :type: smqtk.algorithms.NearestNeighborsIndex
        self.nn_index = plugin.from_plugin_config(
            json_config['nn_index'],
            get_nn_index_impls()
        )

        #: :type: smqtk.algorithms.DescriptorGenerator
        self.descriptor_generator_inst = plugin.from_plugin_config(
            self.json_config['descriptor_generator'],
            get_descriptor_generator_impls()
        )

        @self.route("/count", methods=['GET'])
        def count():
            """
            Return the number of elements represented in this index.
            """
            return flask.jsonify(**{
                "count": self.nn_index.count(),
            })

        @self.route("/compute/<path:uri>", methods=["POST"])
        def compute(uri):
            """
            Compute the descriptor for a URI specified data element using the
            configured descriptor generator.

            See ``compute_nearest_neighbors`` method docstring for URI
            specifications accepted.

            If the a descriptor index was configured and update was turned on,
            we add the computed descriptor to the index.

            JSON Return format::
                {
                    "success": <bool>

                    "message": <str>

                    "descriptor": <None|list[float]>

                    "reference_uri": <str>
                }

            :param uri: URI data specification.

            """
            descriptor = None
            try:
                descriptor = self.generate_descriptor_for_uri(uri)
                message = "Descriptor generated"
                descriptor = list(map(float, descriptor.vector()))
            except ValueError as ex:
                message = "Input value issue: %s" % str(ex)
            except RuntimeError as ex:
                message = "Descriptor extraction failure: %s" % str(ex)

            return flask.jsonify(
                success=descriptor is not None,
                message=message,
                descriptor=descriptor,
                reference_uri=uri,
            )

        @self.route("/nn/<path:uri>")
        @self.route("/nn/n=<int:n>/<path:uri>")
        @self.route("/nn/n=<int:n>/<int:start_i>:<int:end_i>/<path:uri>")
        def compute_nearest_neighbors(uri, n=10, start_i=None, end_i=None):
            """
            Data modes for upload/use:

                - local filepath
                - base64
                - http/s URL
                - existing data/descriptor UUID

            The following sub-sections detail how different URI's can be used.

            Local Filepath
            --------------
            The URI string must be prefixed with ``file://``, followed by the
            full path to the data file to describe.

            Base 64 data
            ------------
            The URI string must be prefixed with "base64://", followed by the
            base64 encoded string. This mode also requires an additional
            ``?content_type=`` to provide data content type information. This
            mode saves the encoded data to temporary file for processing.

            HTTP/S address
            --------------
            This is the default mode when the URI prefix is none of the above.
            This uses the requests module to locally download a data file
            for processing.

            Existing Data/Descriptor by UUID
            --------------------------------
            When given a uri prefixed with "uuid://", we interpret the remainder
            of the uri as the UUID of a descriptor already present in the
            configured descriptor index. If the given UUID is not present in the
            index, a KeyError is raised.

            JSON Return format
            ------------------
                {
                    "success": <bool>

                    "message": <str>

                    "neighbors": <None|list[float]>

                    "reference_uri": <str>
                }

            :param n: Number of neighbors to query for
            :param start_i: The starting index of the neighbor vectors to slice
                into for return.
            :param end_i: The ending index of the neighbor vectors to slice
                into for return.
            :type uri: str

            """
            descriptor = None
            try:
                descriptor = self.generate_descriptor_for_uri(uri)
                message = "descriptor computed"
            except ValueError as ex:
                message = "Input data issue: %s" % str(ex)
            except RuntimeError as ex:
                message = "Descriptor generation failure: %s" % str(ex)

            # Base pagination slicing based on provided start and end indices,
            # otherwise clamp to beginning/ending of queried neighbor sequence.
            page_slice = slice(start_i or 0, end_i or n)
            neighbors = []
            dists = []
            if descriptor is not None:
                try:
                    neighbors, dists = \
                        self.nn_index.nn(descriptor, n)
                except ValueError as ex:
                    message = "Descriptor or index related issue: %s" % str(ex)

            # TODO: Return the optional descriptor vectors for the neighbors
            # noinspection PyTypeChecker
            d = {
                "success": bool(descriptor is not None),
                "message": message,
                "neighbors": [n.uuid() for n in neighbors[page_slice]],
                "distances": dists[page_slice],
                "reference_uri": uri
            }
            return flask.jsonify(d)

    def get_config(self):
        return self.json_config

    def resolve_data_element(self, uri):
        """
        Given the URI to some data, resolve it down to a DataElement instance.

        :raises ValueError: Issue with the given URI regarding either URI source
            resolution or data resolution.

        :param uri: URI to data
        :type uri: str
        :return: DataElement instance wrapping given URI to data.
        :rtype: smqtk.representation.DataElement

        """
        self._log.debug("Resolving URI: %s", uri)
        # Resolve URI into appropriate DataElement instance
        if uri[:7] == "file://":
            self._log.debug("Given local disk filepath")
            filepath = uri[7:]
            if not os.path.isfile(filepath):
                raise ValueError("File URI did not point to an existing file "
                                 "on disk.")
            else:
                de = DataFileElement(filepath)

        elif uri[:9] == "base64://":
            self._log.debug("Given base64 string")
            content_type = flask.request.args.get('content_type', None)
            self._log.debug("Content type: %s", content_type)
            if not content_type:
                raise ValueError("No content-type with given base64 data")
            else:
                b64str = uri[9:]
                de = DataMemoryElement.from_base64(b64str, content_type)

        else:
            self._log.debug("Given URL")
            try:
                de = DataUrlElement(uri)
            except requests.HTTPError as ex:
                raise ValueError("Failed to initialize URL element due to "
                                 "HTTPError: %s" % str(ex))

        return de

    def generate_descriptor_for_uri(self, uri):
        """
        Given the URI to some data, resolve it and compute its descriptor,
        returning a DescriptorElement.

        :param uri: URI to data
        :type uri: str

        :return: DescriptorElement instance of the generate descriptor.
        :rtype: smqtk.representation.DescriptorElement

        """
        # Short-cut if we are given data/descriptor UUID via URI
        if uri[:7] == 'uuid://':
            descriptor = self.descr_index[uri[7:]]
        else:
            de = self.resolve_data_element(uri)
            descriptor = self.descriptor_generator_inst.compute_descriptor(
                de, self.descr_elem_factory
            )
            if self.update_index:
                self._log.info("Updating index with new descriptor")
                self.descr_index.add_descriptor(descriptor)
            if not descriptor.has_vector():
                raise RuntimeError("No descriptor content")
        return descriptor


APPLICATION_CLASS = NearestNeighborServiceServer



"""
Auto-encoder network model definition. Only defined if kwcnn is available.
"""
from .kwcnndescriptor import kwcnn


AutoEncoderModel = None
if kwcnn is not None:
    class AutoEncoderModel(kwcnn.core.KWCNN_Auto_Model):  # NOQA
        """FCNN Model."""

        def __init__(self, *args, **kwargs):
            """FCNN init."""
            super(AutoEncoderModel, self).__init__(*args, **kwargs)
            self.greyscale = kwargs.get("greyscale", False)
            self.bottleneck = kwargs.get("bottleneck", 64)
            self.trimmed = kwargs.get("trimmed", False)

        def _input_shape(self):
            return (64, 64, 1) if self.greyscale else (64, 64, 3)

        # noinspection PyProtectedMember
        def architecture(self, batch_size, in_width, in_height,
                         in_channels, out_classes):
            """FCNN architecture."""
            input_height, input_width, input_channels = self._input_shape()
            _nonlinearity = kwcnn.tpl._lasagne.nonlinearities.LeakyRectify(
                leakiness=(1. / 10.)
            )

            #: :type: list[lasagne.layers.Layer]
            layer_list = [
                kwcnn.tpl._lasagne.layers.InputLayer(
                    shape=(None, input_channels, input_height, input_width)
                )
            ]

            for index in range(2):
                layer_list.append(
                    kwcnn.tpl._lasagne.layers.batch_norm(
                        kwcnn.tpl._lasagne.Conv2DLayer(
                            layer_list[-1],
                            num_filters=16,
                            filter_size=(3, 3),
                            nonlinearity=_nonlinearity,
                            W=kwcnn.tpl._lasagne.init.Orthogonal(),
                        )
                    )
                )

            layer_list.append(
                kwcnn.tpl._lasagne.MaxPool2DLayer(
                    layer_list[-1],
                    pool_size=(2, 2),
                    stride=(2, 2),
                )
            )

            for index in range(3):
                layer_list.append(
                    kwcnn.tpl._lasagne.layers.batch_norm(
                        kwcnn.tpl._lasagne.Conv2DLayer(
                            layer_list[-1],
                            num_filters=32,
                            filter_size=(3, 3),
                            nonlinearity=_nonlinearity,
                            W=kwcnn.tpl._lasagne.init.Orthogonal(),
                        )
                    )
                )

            layer_list.append(
                kwcnn.tpl._lasagne.MaxPool2DLayer(
                    layer_list[-1],
                    pool_size=(2, 2),
                    stride=(2, 2),
                )
            )

            for index in range(2):
                layer_list.append(
                    kwcnn.tpl._lasagne.layers.batch_norm(
                        kwcnn.tpl._lasagne.Conv2DLayer(
                            layer_list[-1],
                            num_filters=32,
                            filter_size=(3, 3),
                            nonlinearity=_nonlinearity,
                            W=kwcnn.tpl._lasagne.init.Orthogonal(),
                        )
                    )
                )

            l_reshape0 = kwcnn.tpl._lasagne.layers.ReshapeLayer(
                layer_list[-1],
                shape=([0], -1),
            )

            if self.trimmed:
                return l_reshape0

            l_bottleneck = kwcnn.tpl._lasagne.layers.DenseLayer(
                l_reshape0,
                num_units=self.bottleneck,
                nonlinearity=kwcnn.tpl._lasagne.nonlinearities.tanh,
                W=kwcnn.tpl._lasagne.init.Orthogonal(),
                name="bottleneck",
            )

            return l_bottleneck



from __future__ import absolute_import, division
from __future__ import print_function, unicode_literals

import collections
from copy import deepcopy
import json
import multiprocessing
import numpy as np
import os
import six
import tempfile

from six.moves import zip

from smqtk.algorithms.nn_index import NearestNeighborsIndex
from smqtk.exceptions import ReadOnlyError
from smqtk.representation import (
    get_data_element_impls,
    get_descriptor_index_impls,
    get_key_value_store_impls,
)
from smqtk.representation.descriptor_element import elements_to_matrix
from smqtk.utils import plugin, merge_dict, metrics

# Requires FAISS bindings
try:
    import faiss
except ImportError:
    faiss = None


class FaissNearestNeighborsIndex (NearestNeighborsIndex):
    """
    Nearest-neighbor computation using the FAISS library.
    """

    @staticmethod
    def gpu_supported():
        """
        :return: If FAISS seems to have GPU support or not.
        :rtype: bool
        """
        # Test if the GPU version is available
        if hasattr(faiss, "StandardGpuResources"):
            return True
        else:
            return False

    @classmethod
    def is_usable(cls):
        # if underlying library is not found, the import above will error
        return faiss is not None

    @classmethod
    def get_default_config(cls):
        """
        Generate and return a default configuration dictionary for this class.
        This will be primarily used for generating what the configuration
        dictionary would look like for this class without instantiating it.

        By default, we observe what this class's constructor takes as
        arguments, turning those argument names into configuration dictionary
        keys. If any of those arguments have defaults, we will add those
        values into the configuration dictionary appropriately. The dictionary
        returned should only contain JSON compliant value types.

        It is not be guaranteed that the configuration dictionary returned
        from this method is valid for construction of an instance of this
        class.

        :return: Default configuration dictionary for the class.
        :rtype: dict

        """
        default = super(FaissNearestNeighborsIndex, cls).get_default_config()

        data_element_default_config = plugin.make_config(
            get_data_element_impls())
        default['index_element'] = data_element_default_config
        default['index_param_element'] = deepcopy(data_element_default_config)

        di_default = plugin.make_config(get_descriptor_index_impls())
        default['descriptor_set'] = di_default

        kvs_default = plugin.make_config(get_key_value_store_impls())
        default['idx2uid_kvs'] = kvs_default
        default['uid2idx_kvs'] = deepcopy(kvs_default)

        return default

    @classmethod
    def from_config(cls, config_dict, merge_default=True):
        """
        Instantiate a new instance of this class given the configuration
        JSON-compliant dictionary encapsulating initialization arguments.

        This method should not be called via super unless and instance of the
        class is desired.

        :param config_dict: JSON compliant dictionary encapsulating
            a configuration.
        :type config_dict: dict

        :param merge_default: Merge the given configuration on top of the
            default provided by ``get_default_config``.
        :type merge_default: bool

        :return: Constructed instance from the provided config.
        :rtype: LSHNearestNeighborIndex

        """
        if merge_default:
            cfg = cls.get_default_config()
            merge_dict(cfg, config_dict)
        else:
            cfg = config_dict

        cfg['descriptor_set'] = plugin.from_plugin_config(
            cfg['descriptor_set'], get_descriptor_index_impls()
        )
        cfg['uid2idx_kvs'] = plugin.from_plugin_config(
            cfg['uid2idx_kvs'], get_key_value_store_impls()
        )
        cfg['idx2uid_kvs'] = plugin.from_plugin_config(
            cfg['idx2uid_kvs'], get_key_value_store_impls()
        )

        if (cfg['index_element'] and
                cfg['index_element']['type']):
            index_element = plugin.from_plugin_config(
                cfg['index_element'], get_data_element_impls())
            cfg['index_element'] = index_element
        else:
            cfg['index_element'] = None

        if (cfg['index_param_element'] and
                cfg['index_param_element']['type']):
            index_param_element = plugin.from_plugin_config(
                cfg['index_param_element'], get_data_element_impls())
            cfg['index_param_element'] = index_param_element
        else:
            cfg['index_param_element'] = None

        return super(FaissNearestNeighborsIndex, cls).from_config(cfg, False)

    def __init__(self, descriptor_set, idx2uid_kvs, uid2idx_kvs,
                 index_element=None, index_param_element=None,
                 read_only=False, factory_string='IVF1,Flat',
                 use_multiprocessing=True, use_gpu=False, gpu_id=0,
                 random_seed=None):
        """
        Initialize FAISS index properties. Does not contain a queryable index
        until one is built via the ``build_index`` method, or loaded from
        existing model files.

        :param descriptor_set: Index in which DescriptorElements will be
            stored.
        :type descriptor_set: smqtk.representation.DescriptorIndex

        :param idx2uid_kvs: Key-value storage mapping FAISS indexed vector
            index to descriptor UID.  This should be the inverse of
            `uid2idx_kvs`.
        :type idx2uid_kvs: smqtk.representation.KeyValueStore

        :param uid2idx_kvs: Key-value storage mapping descriptor UIDs to FAISS
            indexed vector index.  This should be the inverse of `idx2uid_kvs`.
        :type uid2idx_kvs: smqtk.representation.KeyValueStore

        :param index_element: Optional DataElement used to load/store the
            index. When None, the index will only be stored in memory.
        :type index_element: None | smqtk.representation.DataElement

        :param index_param_element: Optional DataElement used to load/store
            the index parameters. When None, the index will only be stored in
            memory.
        :type index_param_element: None | smqtk.representation.DataElement

        :param read_only: If True, `build_index` will error if there is an
            existing index. False by default.
        :type read_only: bool

        :param factory_string: String to pass to FAISS' `index_factory`;
            see the documentation [1] on this feature for more details.
            TODO(john.moeller): Flat indexes are not supported, so set the
            default to 'IVF1,Flat', which is essentially a flat index.
        :type factory_string: str | unicode

        :param use_multiprocessing: Whether or not to use discrete processes
            as the parallelization agent vs python threads.
        :type use_multiprocessing: bool

        :param use_gpu: Use a GPU index if GPU support is available.  A
            RuntimeError is thrown during instance construction if GPU support
            is not available and this flag is true.  See the following for
            FAISS GPU documentation and limitations:

                https://github.com/facebookresearch/faiss/wiki/Faiss-on-the-GPU
        :type use_gpu: bool

        :param gpu_id: If the GPU implementation is available for FAISS
            (automatically determined) use the GPU with this device number /
            ID.
        :type gpu_id: int

        :param random_seed: Integer to use as the random number generator
            seed.
        :type random_seed: int | None

        [1]: https://github.com/facebookresearch/faiss/wiki/High-level-interface-and-auto-tuning#index-factory

        """
        super(FaissNearestNeighborsIndex, self).__init__()

        if not isinstance(factory_string, six.string_types):
            raise ValueError('The factory_string parameter must be a '
                             'recognized string type.')

        self._descriptor_set = descriptor_set
        self._idx2uid_kvs = idx2uid_kvs
        self._uid2idx_kvs = uid2idx_kvs
        self._index_element = index_element
        self._index_param_element = index_param_element
        self.read_only = read_only
        self.factory_string = str(factory_string)
        self.use_multiprocessing = use_multiprocessing
        self._use_gpu = use_gpu
        self._gpu_id = gpu_id
        self.random_seed = None
        if random_seed is not None:
            self.random_seed = int(random_seed)
        # Index value for the next added element.  Reset to 0 on a build.
        self._next_index = 0

        # Place-holder for option GPU resource reference. Just exist for the
        # duration of the index converted with it.
        self._gpu_resources = None
        if self._use_gpu and not self.gpu_supported():
            raise RuntimeError("Requested GPU use but FAISS does not seem to "
                               "support GPU functionality.")

        # Lock for accessing FAISS model components.
        self._model_lock = multiprocessing.RLock()
        # Placeholder for FAISS model instance.
        self._faiss_index = None

        # Load the index/parameters if one exists
        self._load_faiss_model()

    def get_config(self):
        config = {
            "descriptor_set": plugin.to_plugin_config(self._descriptor_set),
            "uid2idx_kvs": plugin.to_plugin_config(self._uid2idx_kvs),
            "idx2uid_kvs": plugin.to_plugin_config(self._idx2uid_kvs),
            "factory_string": self.factory_string,
            "read_only": self.read_only,
            "random_seed": self.random_seed,
            "use_multiprocessing": self.use_multiprocessing,
            "use_gpu": self._use_gpu,
            "gpu_id": self._gpu_id,
        }
        if self._index_element:
            config['index_element'] = plugin.to_plugin_config(
                self._index_element)
        if self._index_param_element:
            config['index_param_element'] = plugin.to_plugin_config(
                self._index_param_element)

        return config

    def _convert_index(self, faiss_index):
        """
        Convert the given index to a GpuIndex if `use_gpu` is True, otherwise
        return the index given (no-op).

        :param faiss_index: Index to convert.
        :type faiss_index: faiss.Index

        :return: Optionally converted index.
        :rtype: faiss.Index | faiss.GpuIndex

        """
        # If we're to use a GPU index and what we're given isn't already a GPU
        # index.
        if self._use_gpu and not isinstance(faiss_index, faiss.GpuIndex):
            self._log.debug("-> GPU-enabling index")
            # New resources
            self._gpu_resources = faiss.StandardGpuResources()
            faiss_index = faiss.index_cpu_to_gpu(faiss.StandardGpuResources(),
                                                 self._gpu_id, faiss_index)
        return faiss_index

    def _index_factory_wrapper(self, d, factory_string):
        """
        Create a FAISS index for the given descriptor dimensionality and
        factory string.

        This *always* produces an index using the L2 metric.

        :param d: Integer indexed vector dimensionality.
        :type d: int

        :param factory_string: Factory string to drive index generation.
        :type factory_string: str

        :return: Constructed index.
        :rtype: faiss.Index | faiss.GpuIndex
        """
        self._log.debug("Creating index by factory: '%s'", factory_string)
        index = faiss.index_factory(d, factory_string, faiss.METRIC_L2)
        return self._convert_index(index)

    def _has_model_data(self):
        """
        Check if configured model files are configured and not empty.
        """
        with self._model_lock:
            return (self._index_element and
                    self._index_param_element and
                    not self._index_element.is_empty() and
                    not self._index_param_element.is_empty())

    def _load_faiss_model(self):
        """
        Load the FAISS model from the configured DataElement
        """
        with self._model_lock:
            if self._has_model_data():
                # Load the binary index
                tmp_fp = self._index_element.write_temp()
                self._faiss_index = self._convert_index(
                    # As of Faiss 1.3.0, only str (not unicode) is
                    # accepted in Python 2.7
                    faiss.read_index(str(tmp_fp))
                )
                self._index_element.clean_temp()

                # Params pickle include the build params + our local state
                # params.
                state = json.loads(self._index_param_element.get_bytes())
                self.factory_string = state["factory_string"]
                self.read_only = state["read_only"]
                self.random_seed = state["random_seed"]
                self.use_multiprocessing = state["use_multiprocessing"]
                self._next_index = state["next_index"]

                # Check that descriptor-set and kvstore instances match up in
                # size.
                assert len(self._descriptor_set) == len(self._uid2idx_kvs) == \
                    len(self._idx2uid_kvs) == self._faiss_index.ntotal, \
                    "Not all of our storage elements agree on size: " \
                    "len(dset, uid2idx, idx2uid, faiss_idx) = " \
                    "(%d, %d, %d, %d)" \
                    % (len(self._descriptor_set), len(self._uid2idx_kvs),
                       len(self._idx2uid_kvs), self._faiss_index.ntotal)

    def _save_faiss_model(self):
        """
        Save the index and parameters to the configured DataElements.
        """
        with self._model_lock:
            # Only write to cache elements if they are both writable.
            writable = (self._index_element and
                        self._index_element.writable() and
                        self._index_param_element and
                        self._index_param_element.writable())
            if writable:
                self._log.debug("Storing index: %s", self._index_element)
                # FAISS wants to write to a file, so make a temp file, then
                # read it in, putting bytes into element.
                fd, fp = tempfile.mkstemp()
                try:
                    # Write function needs a CPU index instance, so bring it
                    # down from the GPU if necessary.
                    if self._use_gpu and isinstance(self._faiss_index,
                                                    faiss.GpuIndex):
                        to_write = faiss.index_gpu_to_cpu(self._faiss_index)
                    else:
                        to_write = self._faiss_index
                    faiss.write_index(to_write, fp)
                    # Use the file descriptor to create the file object.
                    # This avoids reopening the file and will automatically
                    # close the file descriptor on exiting the with block.
                    # fdopen() is required because in Python 2 open() does
                    # not accept a file descriptor.
                    with os.fdopen(fd, 'rb') as f:
                        self._index_element.set_bytes(f.read())
                finally:
                    os.remove(fp)
                # Store index parameters used.
                params = {
                    "factory_string": self.factory_string,
                    "read_only": self.read_only,
                    "random_seed": self.random_seed,
                    "use_multiprocessing": self.use_multiprocessing,
                    "next_index": self._next_index,
                }
                self._index_param_element.set_bytes(json.dumps(params))

    def _build_index(self, descriptors):
        """
        Internal method to be implemented by sub-classes to build the index
        with the given descriptor data elements.

        Subsequent calls to this method should rebuild the current index.
        This method shall not add to the existing index nor raise an exception
        to as to protect the current index.

        :param descriptors: Iterable of descriptor elements to build index
            over.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        """
        if self.read_only:
            raise ReadOnlyError("Cannot modify read-only index.")

        self._log.info("Building new FAISS index")

        # We need to fork the iterator, so stick the elements in a list
        desc_list = list(descriptors)
        data, new_uuids = self._descriptors_to_matrix(desc_list)
        n, d = data.shape
        idx_ids = np.arange(n)  # restart IDs from 0.

        # Build a faiss index but don't internalize it until we have a lock.

        faiss_index = self._index_factory_wrapper(d, self.factory_string)
        # noinspection PyArgumentList
        faiss_index.train(data)
        # TODO(john.moeller): This will raise an exception on flat indexes.
        # There's a solution which involves wrapping the index in an
        # IndexIDMap, but it doesn't work because of a bug in FAISS. So for
        # now we don't support flat indexes.
        # noinspection PyArgumentList
        faiss_index.add_with_ids(data, idx_ids)

        assert faiss_index.d == d, \
            "FAISS index dimension doesn't match data dimension"
        assert faiss_index.ntotal == n, \
            "FAISS index size doesn't match data size"

        with self._model_lock:
            self._faiss_index = faiss_index
            self._log.info("FAISS index has been constructed with %d "
                           "vectors", n)

            self._log.debug("Clearing and adding new descriptor elements")
            self._descriptor_set.clear()
            self._descriptor_set.add_many_descriptors(desc_list)
            assert len(self._descriptor_set) == n, \
                "New descriptor set size doesn't match data size"

            self._uid2idx_kvs.clear()
            self._uid2idx_kvs.add_many(
                dict(zip(new_uuids, idx_ids))
            )
            assert len(self._uid2idx_kvs) == n, \
                "New uid2idx map size doesn't match data size."

            self._idx2uid_kvs.clear()
            self._idx2uid_kvs.add_many(
                dict(zip(idx_ids, new_uuids))
            )
            assert len(self._idx2uid_kvs) == n, \
                "New idx2uid map size doesn't match data size."

            self._next_index = n

            self._save_faiss_model()

    def _update_index(self, descriptors):
        """
        Internal method to be implemented by sub-classes to additively update
        the current index with the one or more descriptor elements given.

        If no index exists yet, a new one should be created using the given
        descriptors.

        :param descriptors: Iterable of descriptor elements to add to this
            index.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        :raises RuntimeError: If a given descriptor is already present in this
            index.  Adding a duplicate descriptor would cause duplicates in
            a nearest-neighbor return (no de-duplication).

        """
        if self.read_only:
            raise ReadOnlyError("Cannot modify read-only index.")

        if self._faiss_index is None:
            self._build_index(descriptors)
            return

        self._log.debug('Updating FAISS index')

        # We need to fork the iterator, so stick the elements in a list
        desc_list = list(descriptors)
        data, new_uuids = self._descriptors_to_matrix(desc_list)
        n, d = data.shape

        with self._model_lock:
            # Assert that new descriptors do not intersect with existing
            # descriptors.
            for uid in new_uuids:
                if uid in self._uid2idx_kvs:
                    raise RuntimeError("Descriptor with UID %s already "
                                       "present in this index.")

            old_ntotal = self.count()

            next_next_index = self._next_index + n
            new_ids = np.arange(self._next_index, next_next_index)
            self._next_index = next_next_index

            assert self._faiss_index.d == d, \
                "FAISS index dimension doesn't match data dimension"
            self._faiss_index.add_with_ids(data, new_ids)
            assert self._faiss_index.ntotal == old_ntotal + n, \
                "New FAISS index size doesn't match old + data size"
            self._log.info("FAISS index has been updated with %d"
                           " new vectors", n)

            self._log.debug("Adding new descriptor elements")
            self._descriptor_set.add_many_descriptors(desc_list)
            assert len(self._descriptor_set) == old_ntotal + n, \
                "New descriptor set size doesn't match old + data size"

            self._uid2idx_kvs.add_many(
                dict(zip(new_uuids, new_ids))
            )
            assert len(self._uid2idx_kvs) == old_ntotal + n, \
                "New uid2idx kvs size doesn't match old + new data size."

            self._idx2uid_kvs.add_many(
                dict(zip(new_ids, new_uuids))
            )
            assert len(self._idx2uid_kvs) == old_ntotal + n, \
                "New idx2uid kvs size doesn't match old + new data size."

            self._save_faiss_model()

    def _remove_from_index(self, uids):
        """
        Internal method to be implemented by sub-classes to partially remove
        descriptors from this index associated with the given UIDs.

        :param uids: Iterable of UIDs of descriptors to remove from this index.
        :type uids: collections.Iterable[collections.Hashable]

        :raises KeyError: One or more UIDs provided do not match any stored
            descriptors.

        """
        if self.read_only:
            raise ReadOnlyError("Cannot modify read-only index.")

        with self._model_lock:
            # Check that provided IDs are present in uid2idx mapping.
            uids_d = collections.deque()
            for uid in uids:
                if uid not in self._uid2idx_kvs:
                    raise KeyError(uid)
                uids_d.append(uid)

            # Remove elements from structures
            # - faiss remove_ids requires a np.ndarray of int64 type.
            rm_idxs = np.asarray([self._uid2idx_kvs[uid] for uid in uids_d],
                                 dtype=np.int64)
            self._faiss_index.remove_ids(rm_idxs)
            self._descriptor_set.remove_many_descriptors(uids_d)
            self._uid2idx_kvs.remove_many(uids_d)
            self._idx2uid_kvs.remove_many(rm_idxs)
            self._save_faiss_model()

    def _descriptors_to_matrix(self, descriptors):
        """
        Extract an (n,d) array with the descriptor vectors in each row,
        and a corresponding list of uuids from the list of descriptors.

        :param descriptors: List descriptor elements to add to this
            index.
        :type descriptors: list[smqtk.representation.DescriptorElement]

        :return: An (n,d) array of descriptors (d-dim descriptors in n
            rows), and the corresponding list of descriptor uuids.
        :rtype: (np.ndarray, list[collections.Hashable])
        """
        new_uuids = [desc.uuid() for desc in descriptors]
        sample_v = descriptors[0].vector()
        n, d = len(new_uuids), sample_v.size
        data = np.empty((n, d), dtype=np.float32)
        elements_to_matrix(
            descriptors, mat=data,
            use_multiprocessing=self.use_multiprocessing,
            report_interval=1.0,
        )
        self._log.info("data shape, type: %s, %s",
                       data.shape, data.dtype)
        self._log.info("# uuids: %d", n)
        return data, new_uuids

    def count(self):
        """
        :return: Number of elements in this index.
        :rtype: int
        """
        with self._model_lock:
            # If we don't have a searchable index we don't actually have
            # anything.
            if self._faiss_index:
                return self._faiss_index.ntotal
            else:
                return 0

    def _nn(self, d, n=1):
        """
        Internal method to be implemented by sub-classes to return the nearest
        `N` neighbors to the given descriptor element.

        When this internal method is called, we have already checked that there
        is a vector in ``d`` and our index is not empty.

        :param d: Descriptor element to compute the neighbors of.
        :type d: smqtk.representation.DescriptorElement

        :param n: Number of nearest neighbors to find.
        :type n: int

        :return: Tuple of nearest N DescriptorElement instances, and a tuple of
            the distance values to those neighbors.
        :rtype: (tuple[smqtk.representation.DescriptorElement], tuple[float])

        """
        q = d.vector()[np.newaxis, :].astype(np.float32)

        self._log.debug("Received query for %d nearest neighbors", n)

        with self._model_lock:
            s_dists, s_ids = self._faiss_index.search(q, n)
            s_dists, s_ids = np.sqrt(s_dists[0, :]), s_ids[0, :]
            uuids = [self._idx2uid_kvs[s_id] for s_id in s_ids]

            descriptors = self._descriptor_set.get_many_descriptors(uuids)

        self._log.debug("Min and max FAISS distances: %g, %g",
                        min(s_dists), max(s_dists))

        descriptors = tuple(descriptors)
        d_vectors = elements_to_matrix(descriptors)
        d_dists = metrics.euclidean_distance(d_vectors, q)

        self._log.debug("Min and max descriptor distances: %g, %g",
                        min(d_dists), max(d_dists))

        order = d_dists.argsort()
        uuids, d_dists = zip(*((uuids[oidx], d_dists[oidx]) for oidx in order))

        self._log.debug("Returning query result of size %g", len(uuids))

        return descriptors, tuple(d_dists)


NN_INDEX_CLASS = FaissNearestNeighborsIndex



import abc

from smqtk.algorithms import SmqtkAlgorithm
from smqtk.utils import check_empty_iterable


class HashIndex (SmqtkAlgorithm):
    """
    Specialized ``NearestNeighborsIndex`` for indexing unique hash codes
    bit-vectors) in memory (numpy arrays) using the hamming distance metric.

    Implementations of this interface cannot be used in place of something
    requiring a ``NearestNeighborsIndex`` implementation due to the speciality
    of this interface.

    Only unique bit vectors should be indexed. The ``nn`` method should not
    return the same bit vector more than once for any query.
    """

    def __len__(self):
        return self.count()

    @staticmethod
    def _empty_iterable_exception():
        """
        Create the exception instance to be thrown when no descriptors are
        provided to ``build_index``/``update_index``.

        :return: ValueError instance to be thrown.
        :rtype: ValueError

        """
        return ValueError("No hash vectors in provided iterable.")

    def build_index(self, hashes):
        """
        Build the index with the given hash codes (bit-vectors).

        Subsequent calls to this method should rebuild the current index.  This
        method shall not add to the existing index nor raise an exception to as
        to protect the current index.

        :raises ValueError: No data available in the given iterable.

        :param hashes: Iterable of descriptor elements to build index
            over.
        :type hashes: collections.Iterable[numpy.ndarray[bool]]

        """
        check_empty_iterable(hashes, self._build_index,
                             self._empty_iterable_exception())

    def update_index(self, hashes):
        """
        Additively update the current index with the one or more hash vectors
        given.

        If no index exists yet, a new one should be created using the given hash
        vectors.

        :raises ValueError: No data available in the given iterable.

        :param hashes: Iterable of numpy boolean hash vectors to add to this
            index.
        :type hashes: collections.Iterable[numpy.ndarray[bool]]

        """
        check_empty_iterable(hashes, self._update_index,
                             self._empty_iterable_exception())

    def remove_from_index(self, hashes):
        """
        Partially remove hashes from this index.

        :param hashes: Iterable of numpy boolean hash vectors to remove from
            this index.
        :type hashes: collections.Iterable[numpy.ndarray[bool]]

        :raises ValueError: No data available in the given iterable.
        :raises KeyError: One or more UIDs provided do not match any stored
            descriptors.

        """
        check_empty_iterable(hashes, self._remove_from_index,
                             self._empty_iterable_exception())

    def nn(self, h, n=1):
        """
        Return the nearest `N` neighbor hash codes as bit-vectors to the given
        hash code bit-vector.

        Distances are in the range [0,1] and are the percent different each
        neighbor hash is from the query, based on the number of bits contained
        in the query (normalized hamming distance).

        :raises ValueError: Current index is empty.

        :param h: Hash code to compute the neighbors of. Should be the same bit
            length as indexed hash codes.
        :type h: numpy.ndarray[bool]

        :param n: Number of nearest neighbors to find.
        :type n: int

        :return: Tuple of nearest N hash codes and a tuple of the distance
            values to those neighbors.
        :rtype: (tuple[numpy.ndarray[bool]], tuple[float])

        """
        # Only check for count because we're no longer dealing with descriptor
        # elements.
        if not self.count():
            raise ValueError("No index currently set to query from!")
        return self._nn(h, n)

    @abc.abstractmethod
    def count(self):
        """
        :return: Number of elements in this index.
        :rtype: int
        """
        pass

    @abc.abstractmethod
    def _build_index(self, hashes):
        """
        Internal method to be implemented by sub-classes to build the index with
        the given hash codes (bit-vectors).

        Subsequent calls to this method should rebuild the current index.  This
        method shall not add to the existing index nor raise an exception to as
        to protect the current index.

        :param hashes: Iterable of descriptor elements to build index
            over.
        :type hashes: collections.Iterable[numpy.ndarray[bool]]

        """

    @abc.abstractmethod
    def _update_index(self, hashes):
        """
        Internal method to be implemented by sub-classes to additively update
        the current index with the one or more hash vectors given.

        If no index exists yet, a new one should be created using the given hash
        vectors.

        :param hashes: Iterable of numpy boolean hash vectors to add to this
            index.
        :type hashes: collections.Iterable[numpy.ndarray[bool]]

        """

    @abc.abstractmethod
    def _remove_from_index(self, hashes):
        """
        Internal method to be implemented by sub-classes to partially remove
        hashes from this index.

        :param hashes: Iterable of numpy boolean hash vectors to remove from
            this index.
        :type hashes: collections.Iterable[numpy.ndarray[bool]]

        :raises KeyError: One or more hashes provided do not match any stored
            hashes.  The index should not be modified.

        """

    @abc.abstractmethod
    def _nn(self, h, n=1):
        """
        Internal method to be implemented by sub-classes to return the nearest
        `N` neighbor hash codes as bit-vectors to the given hash code
        bit-vector.

        Distances are in the range [0,1] and are the percent different each
        neighbor hash is from the query, based on the number of bits contained
        in the query (normalized hamming distance).

        When this internal method is called, we have already checked that our
        index is not empty.

        :param h: Hash code to compute the neighbors of. Should be the same bit
            length as indexed hash codes.
        :type h: numpy.ndarray[bool]

        :param n: Number of nearest neighbors to find.
        :type n: int

        :return: Tuple of nearest N hash codes and a tuple of the distance
            values to those neighbors.
        :rtype: (tuple[numpy.ndarray[bool]], tuple[float])

        """
